import os
import urllib.request
import json
from passlib.hash import pbkdf2_sha512
from urllib.parse import quote, urlencode
from requests_oauthlib import OAuth2Session
from requests_oauthlib.compliance_fixes import facebook_compliance_fix
from phanterpwa.backend.decorators import (
    requires_no_authentication
)
from phanterpwa.backend.dataforms import FieldsDALValidateDictArgs
from phanterpwa.i18n import browser_language

from phanterpwa.gallery.integrationDAL import PhanterpwaGalleryUserImage
from tornado import (
    web
)
from phanterpwa.third_parties.xss import xssescape as E
from phanterpwa.backend.security import (
    Serialize,
    SignatureExpired,
    BadSignature,
    URLSafeSerializer,
)
from datetime import (
    datetime
)
from phanterpwa.helpers import (
    HTML,
    HEAD,
    BODY,
    SCRIPT
)


class Prompt(web.RequestHandler):
    """
        url: url: 'api/oauth/prompt/<social_name>'
    """

    def initialize(self, app_name, projectConfig, DALDatabase, i18nTranslator=None, logger_api=None):
        self.app_name = app_name
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.i18nTranslator = i18nTranslator
        if logger_api:
            self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "phanterpwa-authorization,",
                "cache-control"
            ])
        )
        self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS, POST')
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip
        self.phanterpwa_origin = self.request.headers.get('Referer') or self.projectConfig['BACKEND'][self.app_name]['http_address']

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    def options(self, *args):
        self.set_status(200)
        self.write({"status": "OK"})

    @requires_no_authentication()
    def get(self, *args, **kargs):
        social_name = args[0]
        dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}
        url_base = self.projectConfig['BACKEND'][self.app_name]['http_address']
        if social_name == "google":
            client_id = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_id']
            client_secret = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_secret']
            redirect_uri = '{0}/api/oauth/redirect/{1}'.format(url_base, social_name)
            scope = ['openid', 'https://www.googleapis.com/auth/userinfo.email',
                 'https://www.googleapis.com/auth/userinfo.profile']
            oauth = OAuth2Session(client_id, redirect_uri=redirect_uri,
                              scope=scope)
            authorization_url, state = oauth.authorization_url(
                'https://accounts.google.com/o/oauth2/auth',
                access_type="offline",
                prompt="select_account"
            )
            self.DALDatabase.social_auth.insert(
                social_name=social_name,
                request_state=state,
                client_token=self.phanterpwa_client_token,
                origin=self.phanterpwa_origin or url_base
            )
            self.DALDatabase.commit()
            self.set_status(200)
            return self.write({
                "status": "OK",
                "origin": self.phanterpwa_origin if self.phanterpwa_origin else None,
                "redirect": authorization_url,
                "state": state
            })
        elif social_name == "facebook":
            client_id = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_id']
            client_secret = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_secret']
            redirect_uri = '{0}/api/oauth/redirect/{1}'.format(url_base, social_name)
            oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, scope='email')
            oauth = facebook_compliance_fix(oauth)
            authorization_url, state = oauth.authorization_url(
                'https://www.facebook.com/dialog/oauth',
            )
            self.DALDatabase.social_auth.insert(
                social_name=social_name,
                request_state=state,
                client_token=self.phanterpwa_client_token,
                origin=self.phanterpwa_origin
            )
            self.DALDatabase.commit()
            self.set_status(200)
            return self.write({
                "status": "OK",
                "origin": self.phanterpwa_origin if self.phanterpwa_origin else None,
                "redirect": authorization_url,
                "state": state
            })

        message = "An error occurred while trying to authenticate."
        self.set_status(400)
        return self.write({
            'status': 'Bad Request',
            'code': 400,
            'message': message,
            'i18n': {
                'message': self.T(message)
            }
        })

    def post(self, *args, **kargs):
        dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}
        social_name = args[0]
        state = dict_arguments.get("state")
        db = self.DALDatabase
        q_state = db(
            (db.social_auth.social_name == social_name)
            & (db.social_auth.request_state == state)
        ).select().first()
        if q_state:
            self.set_status(200)
            return self.write(json.loads(q_state.user_credentials))
        else:
            self.set_status(400)
            return self.write({"status": "Bad Request"})


class Redirect(web.RequestHandler):
    """
        url: 'api/oauth/redirect/<social_name>'
    """

    def initialize(self, app_name, projectConfig, DALDatabase, i18nTranslator=None, logger_api=None):
        self.app_name = app_name
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.i18nTranslator = i18nTranslator
        if logger_api:
            self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "phanterpwa-authorization,",
                "cache-control"
            ])
        )
        self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS, POST')
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip
        self.phanterpwa_origin = self.request.headers.get('Origin')

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    def options(self, *args):
        self.set_status(200)
        self.write({"status": "OK"})

    def get(self, *args, **kargs):
        social_name = args[0]
        dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}

        url_base = self.projectConfig['BACKEND'][self.app_name]['http_address']
        if social_name == "google":
            state = dict_arguments.get("state")
            q_state = self.DALDatabase(
                (self.DALDatabase.social_auth.social_name == "google") & (self.DALDatabase.social_auth.request_state == state)).select().first()
            origin = None
            if not q_state or q_state.used is True:
                message = "The authentication request has already been used."
                return self.write({
                    'status': 'Bad Request',
                    'code': 400,
                    'message': message,
                    'i18n': {
                        'message': self.T(message)
                    }
                })
            else:
                self.phanterpwa_client_token = q_state.client_token
                q_state.update_record(used=True)
                origin = q_state.origin
            self.DALDatabase.commit()
            uri = "{0}{1}".format(url_base, self.request.uri)
            client_id = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_id']
            client_secret = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_secret']
            redirect_uri = '{0}/api/oauth/redirect/{1}'.format(url_base, social_name)
            scope = ['openid', 'https://www.googleapis.com/auth/userinfo.email',
                 'https://www.googleapis.com/auth/userinfo.profile']
            oauth = OAuth2Session(client_id, redirect_uri=redirect_uri,
                              scope=scope)
            try:
                token = oauth.fetch_token(
                    'https://accounts.google.com/o/oauth2/token',
                    authorization_response=uri,
                    client_secret=client_secret)
            except Exception as e:
                self.logger_api.warning(e)
                message = "There was a problem trying to authenticate using a google account."
                return self.write({
                    'status': 'Bad Request',
                    'code': 400,
                    'uri': uri,
                    'message': message,
                    'i18n': {
                        'message': self.T(message)
                    }
                })
            else:

                url_consult = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['remote_address']
                googleapi = "{0}?access_token={1}".format(
                    url_consult, quote(token['access_token']))

                try:
                    with urllib.request.urlopen(googleapi) as req:
                        googleapi_user = req.read()
                        googleapi_user = json.loads(googleapi_user)
                except Exception as e:
                    self.logger_api.warning(e)
                    message = "There was a problem trying to load user information on google api."
                    self.set_status(400)
                    return self.write({
                        'status': 'Bad Request',
                        'code': 400,
                        'uri': googleapi,
                        'message': message,
                        'i18n': {
                            'message': self.T(message)
                        }
                    })
                else:
                    email_verified = googleapi_user.get("email_verified", False)
                    if not email_verified:
                        self.logger_api.warning(
                            'LOGIN BY OAUTH METHOD (EMAIL NOT VERIFIED)\n\tsocial_login: {0}\n\tname: {1}\n\temail: {2}\n\temail_verified: {3}\n\tlocale: {4}'.format(
                                social_name,
                                googleapi_user.get("name", False),
                                googleapi_user.get("email", False),
                                googleapi_user.get("email_verified", False),
                                googleapi_user.get("locale", False),
                            )
                        )
                        message = "The google email has not been verified."
                        self.set_status(400)
                        return self.write({
                            'status': 'Bad Request',
                            'code': 400,
                            'message': message,
                            'google_api': googleapi_user,
                            'i18n': {
                                'message': self.T(message)
                            }
                        })
                    email = googleapi_user.get('email', None)

                    if email:
                        q_user = self.DALDatabase(self.DALDatabase.auth_user.email == email).select().first()
                        if q_user:
                            self.logger_api.info(
                                'LOGIN BY OAUTH METHOD\n\tsocial_login: {0}\n\tname: {1}\n\temail: {2}'.format(
                                    social_name,
                                    googleapi_user.get("name", False),
                                    googleapi_user.get("email", False)
                                )
                            )
                            q_user.update_record(activated=True)
                            timeout_token_user = self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire_remember_me']
                            t_user = Serialize(
                                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                timeout_token_user
                            )
                            content = {
                                'id': str(q_user.id),
                                'email': email
                            }
                            token_user = t_user.dumps(content)
                            q_role = self.DALDatabase(
                                (self.DALDatabase.auth_membership.auth_user == q_user.id)
                                & (self.DALDatabase.auth_group.id == self.DALDatabase.auth_membership.auth_group)
                            ).select(self.DALDatabase.auth_group.role, orderby=self.DALDatabase.auth_group.grade)
                            roles = [x.role for x in q_role]
                            role = None
                            if roles:
                                role = roles[-1]
                            q_user.update_record(login_attempts=0)
                            t_client = Serialize(
                                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire']
                            )
                            t_url = URLSafeSerializer(
                                self.projectConfig['BACKEND'][self.app_name]["secret_key"],
                                salt="url_secret_key"
                            )
                            r_client = self.DALDatabase(self.DALDatabase.client.token == self.phanterpwa_client_token).select().first()
                            if r_client:
                                r_client.delete_record()
                            id_client = self.DALDatabase.client.insert(auth_user=q_user.id, date_created=datetime.now())
                            q_client = self.DALDatabase(self.DALDatabase.client.id == id_client).select().first()
                            content = {
                                'id_user': str(q_user.id),
                                'id_client': str(id_client),
                                'user_agent': self.phanterpwa_user_agent,
                                'remote_addr': self.phanterpwa_remote_ip
                            }
                            token_url = t_url.dumps(content)
                            token_client = t_client.dumps(content)
                            q_client.update_record(
                                token=token_client,
                                date_created=datetime.now(),
                                remember_me=True,
                                locked=False,
                            )

                            if not q_user.permit_mult_login:
                                r_client = self.DALDatabase(
                                    (self.DALDatabase.client.auth_user == q_user.id)
                                    & (self.DALDatabase.client.token != token_client)
                                ).select()
                                if r_client:
                                    r_client = self.DALDatabase(
                                        (self.DALDatabase.client.auth_user == q_user.id)
                                        & (self.DALDatabase.client.token != token_client)
                                    ).delete()
                            user_image = PhanterpwaGalleryUserImage(q_user.id, self.DALDatabase, self.projectConfig)
                            social_image = googleapi_user.get("picture", None)

                            redirect = "{0}#_phanterpwa:/oauth/{1}/{2}".format(
                                origin if origin else self.projectConfig['BACKEND'][self.app_name]['http_address'],
                                social_name, state
                            )
                            q_state.update_record(user_credentials=json.dumps({
                                'authorization': token_user,
                                'client_token': token_client,
                                'url_token': token_url,
                                'auth_user': {
                                    'id': str(q_user.id),
                                    'first_name': E(q_user.first_name),
                                    'last_name': E(q_user.last_name),
                                    'email': email,
                                    'remember_me': q_client.remember_me,
                                    'roles': roles,
                                    'role': role,
                                    'activated': True,
                                    'image': user_image.id_image,
                                    'social_image': social_image,
                                    'two_factor': False,
                                    'multiple_login': q_user.permit_mult_login,
                                    'social_login': social_name
                                }
                            }))
                            self.DALDatabase.commit()
                            self.set_status(200)
                            return self.write(
                                str(HTML(HEAD(), BODY(SCRIPT(f"setTimeout(function(){{window.location = '{redirect}';}}, 10)")))))
                        else:
                            self.logger_api.warning(
                                'CREATING ACCOUNT AND LOGIN BY OAUTH METHOD\n\tsocial_login: {0}\n\tname: {1}\n\temail: {2}\n\temail_verified: {3}\n\tlocale: {4}'.format(
                                    social_name,
                                    googleapi_user.get("name", False),
                                    googleapi_user.get("email", False),
                                    googleapi_user.get("email_verified", False),
                                    googleapi_user.get("locale", False),
                                )
                            )
                            new_password = os.urandom(3).hex()
                            password_hash = pbkdf2_sha512.hash("password{0}{1}".format(
                                new_password, self.projectConfig['BACKEND'][self.app_name]['secret_key']))
                            table = self.DALDatabase.auth_user
                            social_image = googleapi_user.get("picture", None)
                            first_name = googleapi_user.get("given_name", "")
                            last_name = googleapi_user.get("family_name", "")
                            dict_arguments = {
                                "first_name": first_name,
                                "last_name": last_name,
                                "email": email,
                                "password_hash": password_hash,
                                "activated": True
                            }
                            result = FieldsDALValidateDictArgs(
                                dict_arguments,
                                *[table[x] for x in table.fields if x in [
                                    "first_name", "last_name", "email", "password_hash"]]
                            )
                            r = result.validate_and_insert(self.DALDatabase.auth_user)
                            if r and r.id:
                                q_user = self.DALDatabase(self.DALDatabase.auth_user.id == r.id).select().first()
                                id_user = q_user.id
                                q_user.update_record(activated=True)
                                if r.id == 1:
                                    role = "root"
                                    id_role = self.DALDatabase(self.DALDatabase.auth_group.role == 'root').select().first()
                                    if id_role:
                                        self.DALDatabase.auth_membership.insert(
                                            auth_user=1,
                                            auth_group=id_role.id
                                        )
                                else:
                                    role = "user"
                                    self.DALDatabase.auth_membership.insert(auth_user=r.id, auth_group=3)
                                t_user = Serialize(
                                    self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                    self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
                                )
                                content_user = {
                                    'id': str(r.id),
                                    'email': dict_arguments['email']
                                }
                                token_user = t_user.dumps(content_user)
                                token_client = self.phanterpwa_client_token
                                id_client = self.DALDatabase.client.update_or_insert(auth_user=r.id)
                                t_client = Serialize(
                                    self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                    self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire']
                                )
                                t_url = URLSafeSerializer(
                                    self.projectConfig['BACKEND'][self.app_name]["secret_key"],
                                    salt="url_secret_key"
                                )
                                content_client = {
                                    'id_user': str(r.id),
                                    'id_client': str(id_client),
                                    'user_agent': self.phanterpwa_user_agent,
                                    'remote_addr': self.phanterpwa_remote_ip
                                }
                                token_url = t_url.dumps(content_client)
                                token_client = t_client.dumps(content_client)
                                q_client = self.DALDatabase(self.DALDatabase.client.id == id_client).select().first()
                                q_client.update_record(
                                    token=token_client,
                                    date_created=datetime.now()
                                )
                                r_client = self.DALDatabase(self.DALDatabase.client.token == self.phanterpwa_client_token).select().first()
                                if r_client:
                                    r_client.delete_record()
                                if not q_user.permit_mult_login:
                                    r_client = self.DALDatabase(
                                        (self.DALDatabase.client.auth_user == id_user)
                                        & (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                    ).select()
                                    if r_client:
                                        r_client = self.DALDatabase(
                                            (self.DALDatabase.client.auth_user == id_user)
                                            & (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                        ).remove()
                                user_image = PhanterpwaGalleryUserImage(r.id, self.DALDatabase, self.projectConfig)
                                self.set_status(201)
                                roles = ["user"]
                                role = "user"
                                redirect = "{0}#_phanterpwa:/oauth/{1}/{2}".format(
                                    origin if origin else self.projectConfig['BACKEND'][self.app_name]['http_address'],
                                    social_name, state
                                )
                                q_state.update_record(user_credentials=json.dumps({
                                    'authorization': token_user,
                                    'client_token': token_client,
                                    'url_token': token_url,
                                    'auth_user': {
                                        'id': str(q_user.id),
                                        'first_name': E(q_user.first_name),
                                        'last_name': E(q_user.last_name),
                                        'email': email,
                                        'remember_me': q_client.remember_me,
                                        'roles': roles,
                                        'role': role,
                                        'activated': True,
                                        'image': user_image.id_image,
                                        'social_image': social_image,
                                        'two_factor': False,
                                        'multiple_login': q_user.permit_mult_login,
                                        'social_login': social_name
                                    }
                                }))
                                self.DALDatabase.commit()
                                return self.write(
                                    str(HTML(HEAD(), BODY(SCRIPT(f"setTimeout(function(){{window.location = '{redirect}';}}, 10)")))))
        elif social_name == "facebook":
            state = dict_arguments.get("state")
            q_state = self.DALDatabase(
                (self.DALDatabase.social_auth.social_name == "facebook") & (self.DALDatabase.social_auth.request_state == state)).select().first()
            origin = None
            if not q_state:
                message = "The authentication request has already been used."
                return self.write({
                    'status': 'Bad Request',
                    'code': 400,
                    'message': message,
                    'i18n': {
                        'message': self.T(message)
                    }
                })
            else:
                self.phanterpwa_client_token = q_state.client_token
                origin = q_state.origin
                q_state.delete_record()
            self.DALDatabase.commit()
            uri = "{0}{1}".format(url_base, self.request.uri)
            client_id = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_id']
            client_secret = self.projectConfig['OAUTH_{0}'.format(social_name.upper())]['client_secret']
            redirect_uri = '{0}/api/oauth/redirect/{1}'.format(url_base, social_name)
            oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, scope='email')
            try:
                token = oauth.fetch_token(
                    'https://graph.facebook.com/oauth/access_token',
                    authorization_response=uri,
                    client_secret=client_secret)
            except Exception as e:
                self.logger_api.warning(e)
                message = "There was a problem trying to authenticate using a facebook account."
                return self.write({
                    'status': 'Bad Request',
                    'code': 400,
                    'uri': uri,
                    'message': message,
                    'i18n': {
                        'message': self.T(message)
                    }
                })
            else:
                try:
                    facebook_user = oauth.get(
                        'https://graph.facebook.com/me?fields=name,first_name,last_name,email,picture')
                    facebookapi_user = json.loads(facebook_user.content)

                except Exception as e:
                    self.logger_api.warning(e)
                    message = "There was a problem trying to load user information on facebook api."
                    self.set_status(400)
                    return self.write({
                        'status': 'Bad Request',
                        'code': 400,
                        'message': message,
                        'i18n': {
                            'message': self.T(message)
                        }
                    })
                else:
                    email = facebookapi_user.get('email', None)
                    social_image = None
                    try:
                        social_image = facebookapi_user["picture"]["data"]["url"]
                    except Exception as e:
                        self.logger_api.warning(e)
                    if email:
                        q_user = self.DALDatabase(self.DALDatabase.auth_user.email == email).select().first()
                        if q_user:
                            q_user.update_record(activated=True)
                            timeout_token_user = self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire_remember_me']
                            t_user = Serialize(
                                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                timeout_token_user
                            )
                            content = {
                                'id': str(q_user.id),
                                'email': email
                            }
                            token_user = t_user.dumps(content)
                            q_role = self.DALDatabase(
                                (self.DALDatabase.auth_membership.auth_user == q_user.id) &
                                (self.DALDatabase.auth_group.id == self.DALDatabase.auth_membership.auth_group)
                            ).select(self.DALDatabase.auth_group.role, orderby=self.DALDatabase.auth_group.grade)
                            roles = [x.role for x in q_role]
                            role = None
                            if roles:
                                role = roles[-1]
                            q_user.update_record(login_attempts=0)
                            t_client = Serialize(
                                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire']
                            )
                            t_url = URLSafeSerializer(
                                self.projectConfig['BACKEND'][self.app_name]["secret_key"],
                                salt="url_secret_key"
                            )
                            r_client = self.DALDatabase(self.DALDatabase.client.token == self.phanterpwa_client_token).select().first()
                            if r_client:
                                r_client.delete_record()
                            id_client = self.DALDatabase.client.insert(auth_user=q_user.id, date_created=datetime.now())
                            q_client = self.DALDatabase(self.DALDatabase.client.id == id_client).select().first()
                            content = {
                                'id_user': str(q_user.id),
                                'id_client': str(id_client),
                                'user_agent': self.phanterpwa_user_agent,
                                'remote_addr': self.phanterpwa_remote_ip
                            }
                            token_url = t_url.dumps(content)
                            token_client = t_client.dumps(content)
                            q_client.update_record(
                                token=token_client,
                                date_created=datetime.now(),
                                remember_me=True,
                                locked=False,
                            )

                            if not q_user.permit_mult_login:
                                r_client = self.DALDatabase(
                                    (self.DALDatabase.client.auth_user == q_user.id) &
                                    (self.DALDatabase.client.token != token_client)
                                ).select()
                                if r_client:
                                    r_client = self.DALDatabase(
                                        (self.DALDatabase.client.auth_user == q_user.id) &
                                        (self.DALDatabase.client.token != token_client)
                                    ).delete()
                            user_image = PhanterpwaGalleryUserImage(q_user.id, self.DALDatabase, self.projectConfig)
                            redirect = "{0}#_phanterpwa:/oauth/{1}/{2}".format(
                                origin if origin else self.projectConfig['BACKEND'][self.app_name]['http_address'],
                                social_name, state
                            )
                            q_state.update_record(user_credentials=json.dumps({
                                    'authorization': token_user,
                                    'client_token': token_client,
                                    'url_token': token_url,
                                    'auth_user': {
                                        'id': str(q_user.id),
                                        'first_name': E(q_user.first_name),
                                        'last_name': E(q_user.last_name),
                                        'email': email,
                                        'remember_me': q_client.remember_me,
                                        'roles': roles,
                                        'role': role,
                                        'activated': True,
                                        'image': user_image.id_image,
                                        'social_image': social_image,
                                        'two_factor': False,
                                        'multiple_login': q_user.permit_mult_login,
                                        'social_login': social_name
                                    }
                                }))
                            self.DALDatabase.commit()
                            self.set_status(200)
                            return self.write(
                                str(HTML(HEAD(), BODY(SCRIPT(f"setTimeout(function(){{window.location = '{redirect}';}}, 10)")))))
                        else:
                            new_password = os.urandom(3).hex()
                            password_hash = pbkdf2_sha512.hash("password{0}{1}".format(
                                new_password, self.projectConfig['BACKEND'][self.app_name]['secret_key']))
                            table = self.DALDatabase.auth_user
                            social_image = googleapi_user.get("picture", None)
                            first_name = googleapi_user.get("given_name", "")
                            last_name = googleapi_user.get("family_name", "")
                            dict_arguments = {
                                "first_name": first_name,
                                "last_name": last_name,
                                "email": email,
                                "password_hash": password_hash,
                                "activated": True
                            }
                            result = FieldsDALValidateDictArgs(
                                dict_arguments,
                                *[table[x] for x in table.fields if x in [
                                    "first_name", "last_name", "email", "password_hash"]]
                            )
                            r = result.validate_and_insert(self.DALDatabase.auth_user)
                            if r and r.id:
                                q_user = self.DALDatabase(self.DALDatabase.auth_user.id == r.id).select().first()
                                id_user = q_user.id
                                q_user.update_record(activated=True)
                                if r.id == 1:
                                    role = "root"
                                    id_role = self.DALDatabase(self.DALDatabase.auth_group.role == 'root').select().first()
                                    if id_role:
                                        self.DALDatabase.auth_membership.insert(auth_user=1,
                                        auth_group=id_role.id)
                                else:
                                    role = "user"
                                    self.DALDatabase.auth_membership.insert(auth_user=r.id, auth_group=3)
                                t_user = Serialize(
                                    self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                    self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
                                )
                                content_user = {
                                    'id': str(r.id),
                                    'email': dict_arguments['email']
                                }
                                token_user = t_user.dumps(content_user)
                                token_client = self.phanterpwa_client_token
                                id_client = self.DALDatabase.client.update_or_insert(auth_user=r.id)
                                t_client = Serialize(
                                    self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                    self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire']
                                )
                                t_url = URLSafeSerializer(
                                    self.projectConfig['BACKEND'][self.app_name]["secret_key"],
                                    salt="url_secret_key"
                                )
                                content_client = {
                                    'id_user': str(r.id),
                                    'id_client': str(id_client),
                                    'user_agent': self.phanterpwa_user_agent,
                                    'remote_addr': self.phanterpwa_remote_ip
                                }
                                token_url = t_url.dumps(content_client)
                                token_client = t_client.dumps(content_client)
                                q_client = self.DALDatabase(self.DALDatabase.client.id == id_client).select().first()
                                q_client.update_record(
                                    token=token_client,
                                    date_created=datetime.now()
                                )
                                r_client = self.DALDatabase(self.DALDatabase.client.token == self.phanterpwa_client_token).select().first()
                                if r_client:
                                    r_client.delete_record()
                                if not q_user.permit_mult_login:
                                    r_client = self.DALDatabase(
                                        (self.DALDatabase.client.auth_user == id_user) &
                                        (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                    ).select()
                                    if r_client:
                                        r_client = self.DALDatabase(
                                            (self.DALDatabase.client.auth_user == id_user) &
                                            (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                        ).remove()
                                user_image = PhanterpwaGalleryUserImage(r.id, self.DALDatabase, self.projectConfig)
                                self.set_status(201)
                                roles = ["user"]
                                role = "user"
                                redirect = "{0}#_phanterpwa:/oauth/{1}/{2}".format(
                                    origin if origin else self.projectConfig['BACKEND'][self.app_name]['http_address'],
                                    social_name, state
                                )
                                q_state.update_record(user_credentials=json.dumps({
                                        'authorization': token_user,
                                        'client_token': token_client,
                                        'url_token': token_url,
                                        'auth_user': {
                                            'id': str(q_user.id),
                                            'first_name': E(q_user.first_name),
                                            'last_name': E(q_user.last_name),
                                            'email': email,
                                            'remember_me': q_client.remember_me,
                                            'roles': roles,
                                            'role': role,
                                            'activated': True,
                                            'image': user_image.id_image,
                                            'social_image': social_image,
                                            'two_factor': False,
                                            'multiple_login': q_user.permit_mult_login,
                                            'social_login': social_name
                                        }
                                    }))
                                self.DALDatabase.commit()
                                return self.write(
                                    str(HTML(HEAD(), BODY(SCRIPT(f"setTimeout(function(){{window.location = '{redirect}';}}, 10)")))))

            
        message = "An error occurred while trying to authenticate."
        self.set_status(400)
        return self.write({
            'status': 'Bad Request',
            'code': 400,
            'message': message,
            'i18n': {
                'message': self.T(message)
            }
        })



import json
import hashlib
from typing import List, Any, Dict
from tornado import (
    web
)
from ..security import (
    Serialize,
    SignatureExpired,
    BadSignature,
    URLSafeSerializer,
)
from phanterpwa.tools import checkbox_bool
from phanterpwa.i18n import browser_language


class DotDict(dict):
    """
    Safe dictionary with attribute-style access, JSON compatible
    and recursive conversion of nested dicts.

    Examples:
        >>> obj = DotDict({"name": "John", "data": {"age": 30}})
        >>> obj.name  # "John"
        >>> obj.data.age  # 30
        >>> isinstance(obj, dict)  # True
        >>> json.dumps(obj)  # Works perfectly
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the dictionary and recursively convert nested dicts
        """
        super().__init__(*args, **kwargs)
        self._convert_nested_dicts()

    def _convert_nested_dicts(self) -> None:
        """
        Recursively convert all nested dicts to DotDict
        and process lists containing dicts
        """
        for key, value in self.items():
            if isinstance(value, dict) and not isinstance(value, DotDict):
                # Convert dict to DotDict
                self[key] = DotDict(value)
            elif isinstance(value, list):
                # Process list - convert dicts inside the list
                self[key] = self._process_list(value)

    def _process_list(self, lst: List) -> List:
        """
        Process a list converting dicts to DotDict
        """
        processed_list = []
        for item in lst:
            if isinstance(item, dict) and not isinstance(item, DotDict):
                processed_list.append(DotDict(item))
            elif isinstance(item, list):
                processed_list.append(self._process_list(item))
            else:
                processed_list.append(item)
        return processed_list

    def __getattr__(self, name: str) -> Any:
        """
        Allow attribute access: obj.name
        """
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Allow attribute assignment: obj.name = value
        """
        if isinstance(value, dict) and not isinstance(value, DotDict):
            value = DotDict(value)
        elif isinstance(value, list):
            value = self._process_list(value)

        self[name] = value

    def __delattr__(self, name: str) -> None:
        """
        Allow attribute deletion: del obj.name
        """
        try:
            del self[name]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def to_dict(self) -> Dict:
        """
        Recursively convert to normal dict - 100% safe for JSON

        Returns:
            Dict: Normal version of dictionary, without DotDict
        """
        result = {}
        for key, value in self.items():
            if isinstance(value, DotDict):
                result[key] = value.to_dict()
            elif isinstance(value, list):
                result[key] = [
                    item.to_dict() if isinstance(item, DotDict) else item for item in value
                ]
            else:
                result[key] = value
        return result

    def copy(self) -> 'DotDict':
        """
        Return a deep copy of the object

        Returns:
            DotDict: Copy of the object
        """
        return DotDict(self.to_dict())

    def update(self, other: Dict = None, **kwargs) -> None:
        """
        Update dictionary with other values, converting nested dicts

        Args:
            other: Another dictionary to update
            **kwargs: Key-value pairs to update
        """
        if other:
            if isinstance(other, dict):
                for key, value in other.items():
                    self[key] = value
            else:
                for key, value in other:
                    self[key] = value

        for key, value in kwargs.items():
            self[key] = value

        # Reconvert nested dicts after update
        self._convert_nested_dicts()

    def get_path(self, path: str, default: Any = None) -> Any:
        """
        Access by path using dots: obj.get_path('user.data.age')

        Args:
            path: Dot path to access nested values
            default: Default value if path doesn't exist

        Returns:
            Any: Found value or default
        """
        keys = path.split('.')
        current = self

        for key in keys:
            if isinstance(current, (dict, DotDict)) and key in current:
                current = current[key]
            else:
                return default

        return current

    def set_path(self, path: str, value: Any) -> None:
        """
        Set value by path using dots: obj.set_path('user.data.age', 30)

        Args:
            path: Dot path to set the value
            value: Value to be set
        """
        keys = path.split('.')
        current = self

        # Navigate to the second to last level
        for key in keys[:-1]:
            if key not in current or not isinstance(current[key], (dict, DotDict)):
                current[key] = DotDict()
            current = current[key]

        # Set value at the last level
        current[keys[-1]] = value

    # Method for JSON compatibility
    def __json__(self) -> Dict:
        """
        Special method that json.dumps() automatically looks for
        """
        return self.to_dict()


class PhanterPWARequestHandler(web.RequestHandler):
    def initialize(self, app_name, projectConfig, DALDatabase, Redis=None, i18nTranslator=None, logger_api=None, AuthActivityNoRelational=None):
        self.app_name = app_name
        self.AuthActivityNoRelational = AuthActivityNoRelational
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.Redis = Redis
        self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T  # translator

        # CORS Headers
        self._set_cors_headers()
        self._final_response_on_success = None

        # Integrated authentication system
        self._initialize_authentication_system()
        self._initialize_csrf_system()

        # Process client headers
        self._process_client_headers()

        self.phanterpwa_redis_cache = checkbox_bool(self.request.headers.get("phanterpwa-redis-cache", True))
        self.dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}
        if self.Redis:
            if self.phanterpwa_redis_cache is False:
                self._redis_invalidate()

    def phanterpwa_current_user_has_role(self, role):
        if self.phanterpwa_current_user_groups:
            if isinstance(role, (list, tuple, set)):
                if isinstance(role, (list, tuple)):
                    role = set(role)
                auth_user_roles = set([x.role for x in self.phanterpwa_current_user_groups])
                if role.intersection(auth_user_roles):
                    return True
                else:
                    return False
            elif isinstance(role, str):
                return self.phanterpwa_current_user_has_role(set([role]))
        return False

    def _set_cors_headers(self):
        # CORS Headers
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-authorization,"
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-redis-cache,",
                "phanterpwa-client-token,",
                "cache-control"
            ])
        )
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS, POST, PUT, DELETE, PATCH, HEAD')

    def _initialize_authentication_system(self):
        """Initialize integrated authentication system"""
        # Tokens
        self.phanterpwa_client_token = None  # Client token
        self.phanterpwa_authorization = None  # Authentication token

        # database data
        self.phanterpwa_current_client = None  # db data client
        self.phanterpwa_current_user = None  # db data user
        self.phanterpwa_current_user_groups = None  # select user groups

        # Verification flags
        self.phanterpwa_client_application_checked = None
        self.phanterpwa_client_token_checked = None
        self.phanterpwa_authorization_checked = None
        self.phanterpwa_user_token_checked = None

    def _initialize_csrf_system(self):
        self.phanterpwa_csrf_token_content = None
        self.phanterpwa_csrf_token_record = None
        self.phanterpwa_form_identify = None

    def _process_client_headers(self):
        """Process PhanterPWA specific headers"""
        # Language
        if self.request.headers.get("phanterpwa-language"):
            self.phanterpwa_language = self.request.headers.get("phanterpwa-language")
        else:
            self.phanterpwa_language = browser_language(self.request.headers.get("Accept-Language"))

        if hasattr(self, 'i18nTranslator'):
            self.i18nTranslator.direct_translation = self.phanterpwa_language

        # Client info
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip

        # Tokens
        self.phanterpwa_application = self.request.headers.get('phanterpwa-application')
        self.phanterpwa_application_version = self.request.headers.get('phanterpwa-application-version')
        self.phanterpwa_client_token = self.request.headers.get('phanterpwa-client-token')
        self.phanterpwa_authorization = self.request.headers.get('phanterpwa-authorization')

    async def GET(self, *args, **kwargs) -> DotDict:
        """GET with optional authentication verification"""
        # By default, doesn't require authentication
        # Specific handlers can override and call check_user_token() if necessary
        pass

    async def POST(self, *args, **kwargs) -> DotDict:
        """POST with optional verification"""
        pass

    async def PATCH(self, *args, **kwargs) -> DotDict:
        """PATCH with optional verification"""
        pass

    async def PUT(self, *args, **kwargs) -> DotDict:
        """PUT with optional verification"""
        pass

    async def DELETE(self, *args, **kwargs) -> DotDict:
        """DELETE with optional verification"""
        pass

    async def OPTIONS(self, *args, **kwargs):
        """OPTIONS for CORS"""
        self.set_status(204)

    async def HEAD(self, *args, **kwargs):
        """HEAD for verification"""
        self.set_status(200)

    async def get(self, *args, **kwargs):
        if self.Redis:
            cache_key = self._generate_cache_key()
            data = self._redis_load(cache_key)
            if data:
                if isinstance(data, (dict, DotDict)):  # ✅ Only add flag if it's JSON
                    data["redis"] = True
                    code = data.get("code", 200)
                    self.set_status(code)
                self._final_response_on_success = data
                self.write(data)
                self.finish()
                return

        data = await self.GET(*args, **kwargs)

        if isinstance(data, (dict, DotDict)):
            code = data.get("code", 200)
            if 200 <= code < 300 and self.Redis:
                cache_key = self._generate_cache_key()
                self._redis_save(cache_key, data, self.projectConfig['BACKEND'][self.app_name].get('default_cache_response_redis_expire', 86400))
            self.set_status(code)
            self.write(data)
        elif data:
            self.write(data)
        self._final_response_on_success = data

    async def post(self, *args, **kwargs):
        data = await self.POST(*args, **kwargs)
        if isinstance(data, (dict, DotDict)):
            code = data.get("code", 200)
            if 200 <= code < 300:
                self._redis_invalidate()
            self.set_status(code)
            self.write(data)
            self._final_response_on_success = data

    async def patch(self, *args, **kwargs):
        data = await self.PATCH(*args, **kwargs)
        if isinstance(data, (dict, DotDict)):
            code = data.get("code", 200)
            if 200 <= code < 300:
                self._redis_invalidate()
            self.set_status(code)
            self.write(data)
            self._final_response_on_success = data

    async def put(self, *args, **kwargs):
        data = await self.PUT(*args, **kwargs)
        if isinstance(data, (dict, DotDict)):
            code = data.get("code", 200)
            if 200 <= code < 300:
                self._redis_invalidate()
            self.set_status(code)
            self.write(data)
            self._final_response_on_success = data

    async def delete(self, *args, **kwargs):
        data = await self.DELETE(*args, **kwargs)
        if isinstance(data, (dict, DotDict)):
            code = data.get("code", 200)
            self.set_status(code)
            if 200 <= code < 300:
                self._redis_invalidate()
            self.write(data)
            self._final_response_on_success = data

    async def options(self, *args, **kwargs):
        self.OPTIONS(*args, **kwargs)

    async def head(self, *args, **kwargs):
        self.HEAD(*args, **kwargs)

    def check_origin(self, origin):
        return True

    def _redis_save(self, key, data, expire=86400) -> bool:
        if self.Redis and self.phanterpwa_redis_cache:
            try:
                if isinstance(data, (bytes, bytearray)):
                    # file cache
                    if expire:
                        self.Redis.setex(key, expire, data)
                    else:
                        self.Redis.set(key, data)
                else:
                    # (DotDict, Dict)
                    data_json = json.dumps(data)
                    if expire:
                        self.Redis.setex(key, expire, data_json)
                    else:
                        self.Redis.set(key, data_json)

                return True
            except Exception as e:
                if self.logger_api:
                    self.logger_api.error(f"❌ Redis save error key({key}): {e}")
                else:
                    print(f"❌ Redis save error key({key}): {e}")
                return False
        return False

    def _redis_load(self, key) -> DotDict:
        if self.Redis and self.phanterpwa_redis_cache:
            try:
                data = self.Redis.get(key)
                if data is None:
                    return None
                try:
                    return DotDict(json.loads(data))
                except (json.JSONDecodeError, UnicodeDecodeError):
                    # its bytes - direct return
                    return data

            except Exception as e:
                if self.logger_api:
                    self.logger_api.error(f"❌ Redis load error key({key}): {e}")
                else:
                    print(f"❌ Redis load error key({key}): {e}")
                return None

    def _redis_invalidate(self):
        """Invalidate all cache of a group"""
        if self.Redis:
            pattern = f"phanterpwa:{self.app_name}:handler:{self.__class__.__name__}:*"
            keys = self.Redis.keys(pattern)
            if keys:
                self.Redis.delete(*keys)

    def _generate_cache_key(self) -> str:
        """
        Simplified version but still very effective
        """
        # Essential elements for uniqueness
        components = [
            self.request.method,
            self.request.path,
        ]

        if self.dict_arguments:
            args_list = []
            # Preserve iteration order of arguments (usually insertion order)
            for key in self.dict_arguments:
                value_str = self.dict_arguments[key]
                args_list.append((key, value_str))
            components.append(json.dumps(args_list, separators=(',', ':')))

        # User context (if authenticated)
        if hasattr(self, 'phanterpwa_current_user') and self.phanterpwa_current_user:
            components.append(f"user:{self.phanterpwa_current_user.id}")

        # Language context
        if hasattr(self, 'phanterpwa_language'):
            components.append(self.phanterpwa_language)
        hash_components = hashlib.md5(":".join(str(c) for c in components).encode()).hexdigest()
        key_string = f"phanterpwa:{self.app_name}:handler:{self.__class__.__name__}:{hash_components}"
        return key_string

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary



from phanterpwa.frontend import helpers
from phanterpwa.frontend import validations
# pragmas

from org.transcrypt.stubs.browser import __pragma__
__pragma__('alias', "jQuery", "$")
__pragma__('skip')

# it is ignored on transcrypt
window = jQuery = console = document = localStorage = String = setTimeout = Date = RegExp = \
    sessionStorage = this = FileReader = JSON = js_undefined = navigator = __new__ = Date = 0

__pragma__('noskip')


INPUT = helpers.XmlConstructor.tagger("input", True)
I = helpers.XmlConstructor.tagger("i")
DIV = helpers.XmlConstructor.tagger("div")
FORM = helpers.XmlConstructor.tagger("form")
SPAN = helpers.XmlConstructor.tagger("span")
XML = helpers.XML
HR = helpers.XmlConstructor.tagger("hr", True)
I18N = helpers.I18N


__pragma__('kwargs')


class Datepickers():

    def __init__(self, target_selector, **parameters):
        self._days = [
            'Sunday',
            'Monday',
            'Tuesday',
            'Wednesday',
            'Thursday',
            'Friday',
            'Saturday'
        ]
        self._ordinaries = [
            "st",
            "nd",
            "rd",
            "th"
        ]
        self._months = [
            'January',
            'February',
            'March',
            'April',
            'May',
            'June',
            'July',
            'August',
            'September',
            'October',
            'November',
            'December'
        ]
        self._times = [
            'hours',
            'minutes',
            'seconds'
        ]
        self.ordinaries = list(self._ordinaries)
        self.namespace = __new__(Date().getTime())
        self.target_selector = target_selector
        self.days = list(self._days)
        self.months = list(self._months)
        self.times = list(self._times)
        self.date_type = "datetime"
        self.debug = window.PhanterPWA.DEBUG
        if "debug" in parameters:
            self.debug = parameters["debug"]
        if "date_type" in parameters:
            self._date_type(parameters['date_type'])
        self.now = __new__(Date())
        self.selected_date = __new__(Date(self.now.getFullYear(), self.now.getMonth(), self.now.getDate()))
        self.current_date = __new__(Date(self.now.getFullYear(), self.now.getMonth(), 1))
        if jQuery(self.target_selector).length > 0:
            if jQuery(self.target_selector)[0].hasAttribute("phanterpwa-datetimepicker-iso"):
                phanterpwa_datetimepicker_iso = jQuery(self.target_selector).attr("phanterpwa-datetimepicker-iso")
                self.selected_date = __new__(Date(phanterpwa_datetimepicker_iso.replace(" ", "T")))
                self.current_date = __new__(Date(self._apply_format("yyyy-MM-01THH:ss:mm")))
        self.format = "yyyy-MM-dd"
        if self.date_type == "datetime":
            self.format = "yyyy-MM-dd HH:ss:mm"
        if "format" in parameters:
            self.format = self._map_escape_str(parameters['format'])

        self.format_in = "yyyy-MM-dd HH:ss:mm"
        if "current_date" in parameters:
            if "format" in parameters:
                self._read_formated(parameters['current_date'])
            else:
                self._selected_date(parameters['current_date'])

        self.onChoice = None
        if "onChoice" in parameters:
            self.onChoice = parameters["onChoice"]

        if "id_input_target" in parameters:
            self.id_input_target = parameters['id_input_target']

            if jQuery(self.id_input_target).length > 0:
                if validations.check_datetime(
                        jQuery(self.id_input_target).val(), self.format, self.date_type):
                    self._read_formated(jQuery(self.id_input_target).val())
                    iso_format = self._apply_format("yyyy-MM-dd HH:ss:mm")
                    jQuery(self.target_selector).attr("phanterpwa-datetimepicker-iso", iso_format)

    def _onChoice(self):
        if callable(self.onChoice):
            iso_format = self._apply_format("yyyy-MM-dd HH:ss:mm")
            jQuery(self.target_selector).attr("phanterpwa-datetimepicker-iso", iso_format)
            __pragma__('jsiter')
            data = {
                'iso': iso_format,
                'formated': self._apply_format(self.format),
            }
            __pragma__('nojsiter')
            self.onChoice(data)
        else:
            iso_format = self._apply_format("yyyy-MM-dd HH:ss:mm")
            jQuery(self.target_selector).attr("phanterpwa-datetimepicker-iso", iso_format)
            if jQuery(self.id_input_target).length > 0:
                jQuery(self.id_input_target).val(self._apply_format(self.format)).focus().trigger("change")
            if self.debug:
                console.info(self._apply_format(self.format))

    # def _patterns_format(self, value):
    #     if len(str(value)) == 1:

    def _sanitize_i18ns(self, value):
        reserved_letters = {
            "d": "&#100;",
            "M": "&#77;",
            "t": "&#116;",
            "o": "&#111;",
            "y": "&#121;",
            "H": "&#72;",
            "m": "&#109;",
            "s": "&#115;"
        }
        for x in reserved_letters.keys():
            if x in value:
                value = value.replace(x, reserved_letters[x])
        return value

    def _map_escape_str(self, value):
        reserved_letters = {
            r"\d": "&#100;",
            r"\M": "&#77;",
            r"\t": "&#116;",
            r"\o": "&#111;",
            r"\y": "&#121;",
            r"\H": "&#72;",
            r"\m": "&#109;",
            r"\s": "&#115;"
        }
        for x in reserved_letters.keys():
            if x in value:
                value = value.replace(x, reserved_letters[x])
        return value

    def _unsanitize_str(self, value):
        htmls_codes = {
            "&#100;": "d",
            "&#77;": "M",
            "&#116;": "t",
            "&#111;": "o",
            "&#121;": "y",
            "&#72;": "H",
            "&#109;": "m",
            "&#115;": "s"
        }
        for x in htmls_codes.keys():
            if x in value:
                value = value.replace(x, htmls_codes[x])
        return value

    def _read_formated(self, value):
        filted_format = self.format
        filted_value = self._sanitize_i18ns(self._map_escape_str(value))
        if "dddd" in filted_format:
            p = "|".join([self._sanitize_i18ns(x) for x in self.days])
            r = "{0}".format(p)
            r = __new__(RegExp(r))
            console.error(r)
            filted_value = filted_value.js_replace(r, "")
            filted_format = filted_format.replace("dddd", "")

        if "ddd" in filted_format:
            p = "|".join([self._sanitize_i18ns(x[:3]) for x in self.days])
            r = "{0}".format(p)
            r = __new__(RegExp(r))
            filted_value = filted_value.js_replace(r, "")
            filted_format = filted_format.replace("ddd", "")

        if "MMMM" in filted_format:
            p = "|".join([self._sanitize_i18ns(x) for x in self.months])
            r = "{0}".format(p)
            r = __new__(RegExp(r))
            filted_value = filted_value.js_replace(r, "")
            filted_format = filted_format.replace("MMMM", "")

        if "MMM" in filted_format:
            p = "|".join([self._sanitize_i18ns(x[:3]) for x in self.months])
            r = "{0}".format(p)
            r = __new__(RegExp(r))
            filted_value = filted_value.js_replace(r, "")
            filted_format = filted_format.replace("MMM", "")

        if "do" in filted_format or "Mo" in filted_format:
            o = [self._sanitize_i18ns(x) for x in self.ordinaries]
            r = "1" + o[0] + "|2" + o[1] + "|3" + o[2] + "|[0-9]{1,2}" + o[3]
            r = __new__(RegExp(r))
            filted_value = filted_value.js_replace(r, "")
            filted_format = filted_format.replace("do", "").replace("Mo", "")

        if "tt" in filted_format:
            r = self._sanitize_i18ns("/PM|AM/")
            r = __new__(RegExp(r))
            filted_value = filted_value.js_replace(r, "")
            filted_format = filted_format.replace("tt", "")

        day = None
        month = None
        year = None
        hour = "00"
        minute = "00"
        second = "00"

        if "dd" in filted_format:
            ini = filted_format.indexOf("dd")
            day = filted_value[ini:ini + 2]
        else:
            raise ValueError("The format must be 'dd' in your pattern")

        if "MM" in filted_format:
            ini = filted_format.indexOf("MM")
            month = filted_value[ini:ini + 2]
        else:
            raise ValueError("The format must be 'MM' in your pattern")

        if "yyyy" in filted_format:
            ini = filted_format.indexOf("yyyy")
            year = filted_value[ini:ini + 4]
        else:
            raise ValueError("The format must be 'yyyy' in your pattern")

        if "HH" in filted_format:
            ini = filted_format.indexOf("HH")
            hour = filted_value[ini:ini + 2]
        else:
            if self.date_type == "datetime":
                raise ValueError("The datetime format must be 'HH' in your pattern")

        if "mm" in filted_format:
            ini = filted_format.indexOf("mm")
            minute = filted_value[ini:ini + 2]
        else:
            if self.date_type == "datetime":
                raise ValueError("The datetime format must be 'mm' in your pattern")

        if "ss" in filted_format:
            ini = filted_format.indexOf("ss")
            second = filted_value[ini:ini + 2]
        else:
            if self.date_type == "datetime":
                raise ValueError("The datetime format must be 'ss' in your pattern")
        self.selected_date = __new__(Date("{0}-{1}-{2}T{3}:{4}:{5}".format(year, month, day, hour, minute, second)))
        self.current_date = __new__(Date("{0}-{1}-01T{2}:{3}:{4}".format(year, month, hour, minute, second)))

    def _apply_format(self, value):
        smonth = self.selected_date.getMonth()
        sday = self.selected_date.getDate()
        sweek = self.selected_date.getDay()
        syear = self.selected_date.getFullYear()
        shour = self.selected_date.getHours()
        sminute = self.selected_date.getMinutes()
        ssecond = self.selected_date.getSeconds()

        if "dddd" in value:
            value = value.replace("dddd", self._sanitize_i18ns(self.days[sweek]))
        if "ddd" in value:
            value = value.replace("ddd", self._sanitize_i18ns(self.days[sweek][0:3]))
        if "dd" in value:
            value = value.replace("dd", self._zfill(sday, 2))
        if "do" in value:
            if sday == 1:
                ordi = self.ordinaries[0]
            elif sday == 2:
                ordi = self.ordinaries[1]
            elif sday == 3:
                ordi = self.ordinaries[2]
            else:
                ordi = self.ordinaries[3]
            nordi = "{0}{1}".format(sday, self._sanitize_i18ns(ordi))
            value = value.replace("do", nordi)

        if "MMMM" in value:
            value = value.replace("MMMM", self._sanitize_i18ns(self.months[smonth]))
        if "MMM" in value:
            value = value.replace("MMM", self._sanitize_i18ns(self.months[smonth][0:3]))
        if "MM" in value:
            value = value.replace("MM", self._zfill(smonth + 1, 2))
        if "Mo" in value:
            if (smonth + 1) == 1:
                ordi = self.ordinaries[0]
            elif (smonth + 1) == 2:
                ordi = self.ordinaries[1]
            elif (smonth + 1) == 3:
                ordi = self.ordinaries[2]
            else:
                ordi = self.ordinaries[3]
            nordi = "{0}{1}".format(smonth + 1, self._sanitize_i18ns(ordi))
            value = value.replace("Mo", nordi)

        if "yyyy" in value:
            value = value.replace("yyyy", syear)

        if "HH" in value:
            value = value.replace("HH", self._zfill(shour, 2))

        if "mm" in value:
            value = value.replace("mm", self._zfill(sminute, 2))

        if "ss" in value:
            value = value.replace("ss", self._zfill(ssecond, 2))

        AM_PM = "PM"
        if shour < 12:
            AM_PM = "AM"

        if "tt" in value:
            value = value.replace("tt", AM_PM)
        return self._unsanitize_str(value)

    def _zfill(self, number, size):
        number = int(number)
        number = str(number)
        s = number
        for x in range(size - len(number)):
            s = "0" + s
        return s

    def _date_type(self, value):
        if value != "datetime" and value != "date":
            raise ValueError(
                "The date_type must be 'datetime' or 'date'. Given '{0}'".format(value)
            )
        self.date_type = value

    def _selected_date(self, value):
        if self.date_type == "date":
            REGEX_BODY = __pragma__(
                'js',
                '{}',
                r'/^([0-9]{4}-[0-9]{2}-[0-9]{2})/'
            )
            if REGEX_BODY.test(value):
                self.selected_date = __new__(Date("{0}T00:00:00".format(value[0:10])))
                self.current_date = __new__(Date("{0}01T00:00:00".format(value[0:8])))
            else:
                raise ValueError("The current_date must be 'yyyy-mm-dd'. Example: 2019-01-01")

        elif self.date_type == "datetime":
            REGEX_BODY = __pragma__(
                'js',
                '{}',
                r'/^([0-9]{4}-[0-9]{2}-[0-9]{2}.{1}[0-9]{2}:[0-9]{2}:[0-9]{2})/'
            )
            if REGEX_BODY.test(value):
                self.selected_date = __new__(Date("{0}T{1}".format(value[0:10], value[11:19])))
                self.current_date = __new__(Date("{0}01T{1}".format(value[0:8], value[11:19])))
            else:
                raise ValueError("The current_date must be 'yyyy-mm-dd'. Example: 2019-01-01")
        else:
            raise ValueError(
                "The date_type must be 'datetime' or 'date'. Given '{0}'".format(value)
            )

    def previus_year(self):
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        cmonth = self.current_date.getMonth()
        cday = self.current_date.getDate()
        cyear = self.current_date.getFullYear()
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear - 1,
            self._zfill(cmonth + 1, 2),
            self._zfill(cday, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def next_year(self):
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        cmonth = self.current_date.getMonth()
        cday = self.current_date.getDate()
        cyear = self.current_date.getFullYear()
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear + 1,
            self._zfill(cmonth + 1, 2),
            self._zfill(cday, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def previus_month(self):
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        cmonth = self.current_date.getMonth()
        cyear = self.current_date.getFullYear()
        calc_month = cmonth - 1
        if calc_month < 0:
            calc_month = 11
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(calc_month + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def next_month(self):
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        cmonth = self.current_date.getMonth()
        cyear = self.current_date.getFullYear()
        calc_month = cmonth + 1
        if calc_month > 11:
            calc_month = 0
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(calc_month + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def previus_hour(self):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        calc_hour = chour - 1
        if calc_hour < 0:
            calc_hour = 23
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(calc_hour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(calc_hour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def next_hour(self):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        calc_hour = chour + 1
        if calc_hour > 23:
            calc_hour = 0
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(calc_hour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(calc_hour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def previus_minute(self):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        calc_minute = cminute - 1
        if calc_minute < 0:
            calc_minute = 59
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(chour, 2),
            self._zfill(calc_minute, 2),
            self._zfill(csecond, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(calc_minute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def next_minute(self):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        calc_minute = cminute + 1
        if calc_minute > 59:
            calc_minute = 0
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(chour, 2),
            self._zfill(calc_minute, 2),
            self._zfill(csecond, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(calc_minute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def previus_second(self):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        calc_second = csecond - 1
        if calc_second < 0:
            calc_second = 59
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(calc_second, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(calc_second, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def next_second(self):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        calc_second = csecond + 1
        if calc_second > 59:
            calc_second = 0
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(calc_second, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(calc_second, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def set_year(self, value):
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        REGEX_BODY = __pragma__(
            'js',
            '{}',
            r'/^([0-9]{4})/'
        )
        if REGEX_BODY.test(value):
            cmonth = self.current_date.getMonth()
            new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
                value,
                self._zfill(cmonth + 1, 2),
                self._zfill(1, 2),
                self._zfill(chour, 2),
                self._zfill(cminute, 2),
                self._zfill(csecond, 2),
            )
            self.current_date = __new__(Date(new_date))
            self.start()

    def set_month(self, value):
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        cyear = self.current_date.getFullYear()
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(int(value) + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def set_hour(self, value):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(int(value), 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(int(value), 2),
            self._zfill(cminute, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def set_minute(self, value):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        csecond = self.current_date.getSeconds()
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(chour, 2),
            self._zfill(value, 2),
            self._zfill(csecond, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(value, 2),
            self._zfill(csecond, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def set_second(self, value):
        cyear = self.current_date.getFullYear()
        cmonth = self.current_date.getMonth()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        sdate = self.selected_date.toJSON()
        snew_date = "{0}T{1}:{2}:{3}".format(
            sdate[0:10],
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(value, 2),
        )
        self.selected_date = __new__(Date(snew_date))
        new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
            cyear,
            self._zfill(cmonth + 1, 2),
            self._zfill(1, 2),
            self._zfill(chour, 2),
            self._zfill(cminute, 2),
            self._zfill(value, 2),
        )
        self.current_date = __new__(Date(new_date))
        self.start()

    def set_selected(self, jsonDate):
        new_date = __new__(Date(jsonDate.replace(" ", "T")))
        self.selected_date = new_date
        self._onChoice()
        self.close()

    def show_months(self):
        container = DIV(
            _class='phanterpwa_datetimepicker_monthlist_container'
        )

        def _month(el):
            m = jQuery(el).attr('phanterpwa_datetimepicker_month')
            self.set_month(m)
        cmonth = self.current_date.getMonth()
        row_container = DIV(_class="phanterpwa_datetimepicker_row")
        for x in range(12):
            add_class = ""
            if cmonth == x:
                add_class = " selected"
            row_container.append(
                DIV(
                    I18N(self.months[x]),
                    _class="phanterpwa_datetimepicker_monthlist_month phanterpwa_datetimepicker_button{0}".format(
                        add_class
                    ),
                    _phanterpwa_datetimepicker_month=x

                )
            )
        container.append(row_container)
        # jQuery('#phanterpwa_datetimepicker_calendar_{0}'.format(self.namespace)).html("").append(container.xml())
        container.html_to('#phanterpwa_datetimepicker_calendar_{0}'.format(self.namespace))
        jQuery(
            '#phanterpwa_datetimepicker_calendar_{0} .phanterpwa_datetimepicker_monthlist_month'.format(self.namespace)
        ).on(
            'click',
            lambda: _month(this)
        )

    def show_years(self, inicial_year=None):
        container = DIV(
            _class='phanterpwa_datetimepicker_yearlist_container'
        )

        def _year(el):
            m = jQuery(el).attr('phanterpwa_datetimepicker_year')
            self.set_year(m)

        row_container = DIV(_class="phanterpwa_datetimepicker_row")
        cyear = self.current_date.getFullYear()
        if inicial_year is None:
            inicial_year = cyear - 17
        final_year = inicial_year + 35
        for x in range(inicial_year, final_year):
            add_class = ""
            if cyear == x:
                add_class = " selected"
            row_container.append(
                DIV(
                    x,
                    _class="phanterpwa_datetimepicker_yearlist_year phanterpwa_datetimepicker_button{0}".format(
                        add_class
                    ),
                    _phanterpwa_datetimepicker_year=x

                )
            )
        container.append(row_container)
        # jQuery('#phanterpwa_datetimepicker_month_and_calendar_{0}'.format(
        #     self.namespace)).html("").append(container.xml())
        container.html_to('#phanterpwa_datetimepicker_month_and_calendar_{0}'.format(
            self.namespace))
        jQuery(
            '#phanterpwa_datetimepicker_month_and_calendar_{0} .phanterpwa_datetimepicker_yearlist_year'.format(
                self.namespace)
        ).on(
            'click',
            lambda: _year(this)
        )
        jQuery(
            '#phanterpwa_datetimepicker_prev_year_{0}'.format(self.namespace)
        ).html(
            DIV(I(_class="fas fa-angle-double-left"), _class="phanterpwa_datetimepicker_button").xml()
        ).off(
            "click.phanterpwa_datetimepicker_prev_year"
        ).on(
            "click.phanterpwa_datetimepicker_prev_year",
            lambda: self.show_years(inicial_year - 35)
        )
        jQuery(
            '#phanterpwa_datetimepicker_next_year_{0}'.format(self.namespace)
        ).html(
            DIV(I(_class="fas fa-angle-double-right"), _class="phanterpwa_datetimepicker_button").xml()
        ).off(
            "click.phanterpwa_datetimepicker_next_year"
        ).on(
            "click.phanterpwa_datetimepicker_next_year",
            lambda: self.show_years(inicial_year + 35)
        )

    def show_hours(self):
        container = DIV(
            _class='phanterpwa_datetimepicker_hourlist_container'
        )

        def _hour(el):
            m = jQuery(el).attr('phanterpwa_datetimepicker_hour')
            self.set_hour(m)
        chour = self.current_date.getHours()
        row_container = DIV(_class="phanterpwa_datetimepicker_row")
        for x in range(24):
            add_class = ""
            if chour == x:
                add_class = " selected"
            row_container.append(
                DIV(
                    self._zfill(x, 2),
                    _class="phanterpwa_datetimepicker_hourlist_hour phanterpwa_datetimepicker_button{0}".format(
                        add_class
                    ),
                    _phanterpwa_datetimepicker_hour=x

                )
            )
        container.append(row_container)
        jQuery('#phanterpwa_datetimepicker_calendar_month_years_{0}'.format(
            self.namespace)).html("").append(container.xml())
        jQuery(
            '#phanterpwa_datetimepicker_calendar_month_years_{0} .phanterpwa_datetimepicker_hourlist_hour'.format(
                self.namespace
            )
        ).on(
            'click',
            lambda: _hour(this)
        )

    def show_minutes(self):
        container = DIV(
            _class='phanterpwa_datetimepicker_minutelist_container'
        )

        def _minute(el):
            m = jQuery(el).attr('phanterpwa_datetimepicker_minute')
            self.set_minute(m)
        cminute = self.current_date.getMinutes()
        row_container = DIV(_class="phanterpwa_datetimepicker_row")
        for x in range(60):
            add_class = ""
            if cminute == x:
                add_class = " selected"
            row_container.append(
                DIV(
                    self._zfill(x, 2),
                    _class="phanterpwa_datetimepicker_minutelist_minute phanterpwa_datetimepicker_button{0}".format(
                        add_class
                    ),
                    _phanterpwa_datetimepicker_minute=x

                )
            )
        container.append(row_container)
        # jQuery('#phanterpwa_datetimepicker_calendar_month_years_{0}'.format(
        #     self.namespace)).html("").append(container.xml())
        container.html_to('#phanterpwa_datetimepicker_calendar_month_years_{0}'.format(
            self.namespace))
        jQuery(
            '#phanterpwa_datetimepicker_calendar_month_years_{0} .phanterpwa_datetimepicker_minutelist_minute'.format(
                self.namespace)
        ).on(
            'click',
            lambda: _minute(this)
        )

    def show_seconds(self):
        container = DIV(
            _class='phanterpwa_datetimepicker_secondlist_container'
        )

        def _second(el):
            m = jQuery(el).attr('phanterpwa_datetimepicker_second')
            self.set_second(m)
        csecond = self.current_date.getSeconds()
        row_container = DIV(_class="phanterpwa_datetimepicker_row")
        for x in range(60):
            add_class = ""
            if csecond == x:
                add_class = " selected"
            row_container.append(
                DIV(
                    self._zfill(x, 2),
                    _class="phanterpwa_datetimepicker_secondlist_second phanterpwa_datetimepicker_button{0}".format(
                        add_class
                    ),
                    _phanterpwa_datetimepicker_second=x

                )
            )
        container.append(row_container)
        # jQuery('#phanterpwa_datetimepicker_calendar_month_years_{0}'.format(
        #     self.namespace)).html("").append(container.xml())
        container.html_to('#phanterpwa_datetimepicker_calendar_month_years_{0}'.format(
            self.namespace))
        jQuery(
            '#phanterpwa_datetimepicker_calendar_month_years_{0} .phanterpwa_datetimepicker_secondlist_second'.format(
                self.namespace)
        ).on(
            'click',
            lambda: _second(this)
        )

    def close(self):
        jQuery(self.target_selector).find(".phanterpwa-fixed-fulldisplay").removeClass("enabled")

        def delete_datepicker():
            if not jQuery(self.target_selector).find(".phanterpwa-fixed-fulldisplay").hasClass("enabled"):
                jQuery(self.target_selector).find(".phanterpwa-fixed-fulldisplay").remove()
        setTimeout(
            delete_datepicker,
            500
        )

    def start(self):
        cont = 0
        cmonth = self.current_date.getMonth()
        cday = self.current_date.getDate()
        cweek = self.current_date.getDay()
        cyear = self.current_date.getFullYear()
        chour = self.current_date.getHours()
        cminute = self.current_date.getMinutes()
        csecond = self.current_date.getSeconds()
        smonth = self.selected_date.getMonth()
        sday = self.selected_date.getDate()
        sweek = self.selected_date.getDay()
        syear = self.selected_date.getFullYear()
        shour = self.selected_date.getHours()
        sminute = self.selected_date.getMinutes()
        ssecond = self.selected_date.getSeconds()
        nmonth = self.now.getMonth()
        nday = self.now.getDate()
        nyear = self.now.getFullYear()
        AM_PM = "PM"
        if shour < 12:
            AM_PM = "AM"

        container = DIV(
            _id='phanterpwa_datetimepicker_{0}'.format(self.namespace),
            _class='phanterpwa_datetimepicker_wrapper',
        )
        container_calendar_month_years = DIV(
            DIV(
                DIV(
                    DIV(
                        I(_class="fas fa-angle-left"),
                        _class="phanterpwa_datetimepicker_button"),
                    _class='phanterpwa_datetimepicker_previusnext',
                    _id='phanterpwa_datetimepicker_prev_year_{0}'.format(self.namespace)
                ),
                DIV(
                    DIV(
                        cyear,
                        _class='phanterpwa_datetimepicker_button',
                        _id='phanterpwa_datetimepicker_year_{0}'.format(self.namespace)
                    ),
                    _class='phanterpwa_datetimepicker_year',
                ),
                DIV(
                    DIV(
                        I(_class="fas fa-angle-right"),
                        _class="phanterpwa_datetimepicker_button"),
                    _class='phanterpwa_datetimepicker_previusnext',
                    _id='phanterpwa_datetimepicker_next_year_{0}'.format(self.namespace)
                ),
                _class='phanterpwa_datetimepicker_row'
            ),
            _id='phanterpwa_datetimepicker_calendar_month_years_{0}'.format(self.namespace),
        )
        hour_container = DIV(
            DIV(
                HR(),

                DIV(
                    DIV(
                        DIV(
                            I(_class="fas fa-angle-left"),
                            _class="phanterpwa_datetimepicker_button"),
                        _id='phanterpwa_datetimepicker_prev_hour_{0}'.format(self.namespace),
                        _class='phanterpwa_datetimepicker_previusnext'
                    ),
                    DIV(
                        DIV(
                            self._zfill(chour, 2),
                            _class='phanterpwa_datetimepicker_button',
                            _id='phanterpwa_datetimepicker_hour_{0}'.format(self.namespace)
                        ),
                        _class='phanterpwa_datetimepicker_hour',
                    ),
                    DIV(
                        DIV(
                            I(_class="fas fa-angle-right"),
                            _class="phanterpwa_datetimepicker_button"),
                        _id='phanterpwa_datetimepicker_next_hour_{0}'.format(self.namespace),
                        _class='phanterpwa_datetimepicker_previusnext'
                    ),
                    _class="phanterpwa_datetimepicker_time_col"
                ),
                DIV(
                    DIV(
                        DIV(
                            I(_class="fas fa-angle-left"),
                            _class="phanterpwa_datetimepicker_button"),
                        _id='phanterpwa_datetimepicker_prev_minute_{0}'.format(self.namespace),
                        _class='phanterpwa_datetimepicker_previusnext'
                    ),
                    DIV(
                        DIV(
                            self._zfill(cminute, 2),
                            _class='phanterpwa_datetimepicker_button',
                            _id='phanterpwa_datetimepicker_minute_{0}'.format(self.namespace)
                        ),
                        _class='phanterpwa_datetimepicker_minute',
                    ),
                    DIV(
                        DIV(
                            I(_class="fas fa-angle-right"),
                            _class="phanterpwa_datetimepicker_button"),
                        _id='phanterpwa_datetimepicker_next_minute_{0}'.format(self.namespace),
                        _class='phanterpwa_datetimepicker_previusnext'
                    ),
                    _class="phanterpwa_datetimepicker_time_col"
                ),
                DIV(
                    DIV(
                        DIV(
                            I(_class="fas fa-angle-left"),
                            _class="phanterpwa_datetimepicker_button"),
                        _id='phanterpwa_datetimepicker_prev_second_{0}'.format(self.namespace),
                        _class='phanterpwa_datetimepicker_previusnext'
                    ),
                    DIV(
                        DIV(
                            self._zfill(csecond, 2),
                            _class='phanterpwa_datetimepicker_button',
                            _id='phanterpwa_datetimepicker_second_{0}'.format(self.namespace)
                        ),
                        _class='phanterpwa_datetimepicker_second',
                    ),
                    DIV(
                        DIV(
                            I(_class="fas fa-angle-right"),
                            _class="phanterpwa_datetimepicker_button"),
                        _id='phanterpwa_datetimepicker_next_second_{0}'.format(self.namespace),
                        _class='phanterpwa_datetimepicker_previusnext'
                    ),
                    _class="phanterpwa_datetimepicker_time_col"
                ),
                DIV(
                    self.times[0],
                    _class="phanterpwa_datetimepicker_time_col label"
                ),
                DIV(
                    self.times[1],
                    _class="phanterpwa_datetimepicker_time_col label"
                ),
                DIV(
                    self.times[2],
                    _class="phanterpwa_datetimepicker_time_col label"
                ),
                _class="phanterpwa_datetimepicker_row"
            ),
            _class="phanterpwa_datetimepicker_time_container"
        )
        calendar_and_month = DIV(
            DIV(
                DIV(
                    DIV(
                        I(_class="fas fa-angle-left"),
                        _class="phanterpwa_datetimepicker_button"),
                    _id='phanterpwa_datetimepicker_prev_month_{0}'.format(self.namespace),
                    _class='phanterpwa_datetimepicker_previusnext'
                ),
                DIV(
                    DIV(
                        I18N(self.months[cmonth]),
                        _class='phanterpwa_datetimepicker_button',
                        _id='phanterpwa_datetimepicker_month_{0}'.format(self.namespace)
                    ),
                    _class='phanterpwa_datetimepicker_month',
                ),
                DIV(
                    DIV(
                        I(_class="fas fa-angle-right"),
                        _class="phanterpwa_datetimepicker_button"),
                    _id='phanterpwa_datetimepicker_next_month_{0}'.format(self.namespace),
                    _class='phanterpwa_datetimepicker_previusnext'
                ),
                _class='phanterpwa_datetimepicker_row'
            ),
            _id='phanterpwa_datetimepicker_month_and_calendar_{0}'.format(self.namespace)
        )
        container_calendar = DIV(
            DIV(
                DIV(
                    DIV(I18N(self.days[0][:3]), _class="phanterpwa_datetimepicker_unit"),
                    _class='phanterpwa_datetimepicker_weekday phanterpwa_datetimepicker_Sunday'
                ),
                DIV(
                    DIV(I18N(self.days[1][:3]), _class="phanterpwa_datetimepicker_unit"),
                    _class='phanterpwa_datetimepicker_weekday phanterpwa_datetimepicker_Monday'
                ),
                DIV(
                    DIV(I18N(self.days[2][:3]), _class="phanterpwa_datetimepicker_unit"),
                    _class='phanterpwa_datetimepicker_weekday phanterpwa_datetimepicker_Tuesday'
                ),
                DIV(
                    DIV(I18N(self.days[3][:3]), _class="phanterpwa_datetimepicker_unit"),
                    _class='phanterpwa_datetimepicker_weekday phanterpwa_datetimepicker_Wednesday'
                ),
                DIV(
                    DIV(I18N(self.days[4][:3]), _class="phanterpwa_datetimepicker_unit"),
                    _class='phanterpwa_datetimepicker_weekday phanterpwa_datetimepicker_Thursday'
                ),
                DIV(
                    DIV(I18N(self.days[5][:3]), _class="phanterpwa_datetimepicker_unit"),
                    _class='phanterpwa_datetimepicker_weekday phanterpwa_datetimepicker_Friday'
                ),
                DIV(
                    DIV(I18N(self.days[6][:3]), _class="phanterpwa_datetimepicker_unit"),
                    _class='phanterpwa_datetimepicker_weekday phanterpwa_datetimepicker_Saturday'
                ),
                _class='phanterpwa_datetimepicker_row',
            ),
            _id='phanterpwa_datetimepicker_calendar_{0}'.format(self.namespace),
        )
        row_days = DIV(_class='phanterpwa_datetimepicker_row')
        for y in range(cweek):
            row_days.append(
                DIV(_class="phanterpwa_datetimepicker_day_empty phanterpwa_datetimepicker_{0}".format(self.days[y]))
            )
        inicial_day = 1
        for x in range(40):
            if inicial_day < 32:
                cont += 1
                new_date = "{0}-{1}-{2}T{3}:{4}:{5}".format(
                    cyear,
                    self._zfill(cmonth + 1, 2),
                    self._zfill(inicial_day, 2),
                    self._zfill(chour, 2),
                    self._zfill(cminute, 2),
                    self._zfill(csecond, 2),
                )
                d = __new__(Date(new_date))
                if (d.getMonth() == cmonth) and (d.toJSON() is not None):
                    add_class = ""
                    if (d.getMonth() == nmonth) and (d.getFullYear() == nyear) and (d.getDate() == nday):
                        add_class += " phanterpwa_datetimepicker_its_now"
                    if self.selected_date is not None:
                        if (d.getMonth() == smonth) and (d.getFullYear() == syear) and (d.getDate() == sday):
                            add_class += " phanterpwa_datetimepicker_selected"

                    row_days.append(
                        DIV(
                            DIV(
                                d.getDate(),
                                _id="phanterpwa_datetimepicker_unit_{0}".format(self.namespace),
                                _class="phanterpwa_datetimepicker_unit",
                                _phanterpwa_datetimepicker_date=new_date
                            ),
                            _class="phanterpwa_datetimepicker_day phanterpwa_datetimepicker_{0}{1}".format(
                                self._days[d.getDay()],
                                add_class
                            )
                        )
                    )
            inicial_day += 1
        container_calendar.append(row_days)
        calendar_and_month.append(container_calendar)
        container_calendar_month_years.append(calendar_and_month)
        container.append(container_calendar_month_years)
        container.append(hour_container)
        summary = DIV(
            DIV(
                I(_class="fas fa-times"),
                _class='phanterpwa-models-close'),
            DIV(
                DIV(
                    SPAN(self._zfill(sday, 2)), ", ", I18N(self.months[smonth]),
                    _id="phanterpwa_datetimepicker_summary_day_and_month_{0}".format(self.namespace),
                    _class="phanterpwa_datetimepicker_summary_day_and_month"
                ),
                DIV(
                    I18N(self.days[sweek]),
                    _id="phanterpwa_datetimepicker_summary_week_{0}".format(self.namespace),
                    _class="phanterpwa_datetimepicker_summary_week"
                ),
                DIV(
                    syear,
                    _id="phanterpwa_datetimepicker_summary_year_{0}".format(self.namespace),
                    _class="phanterpwa_datetimepicker_summary_year"
                ),
                DIV(
                    "{0}:{1}:{2} {3}".format(
                        self._zfill(shour, 2),
                        self._zfill(sminute, 2),
                        self._zfill(ssecond, 2),
                        AM_PM
                    ),
                    _id="phanterpwa_datetimepicker_summary_hour_{0}".format(self.namespace),
                    _class="phanterpwa_datetimepicker_summary_hour"
                ),
                _class="phanterpwa_datetimepicker_row"
            ),
            DIV(
                DIV(
                    DIV(
                        _id="phanterpwa_datetimepicker_summary_current_data_{0}".format(self.namespace),
                    ),
                    _class="phanterpwa_datetimepicker_row"
                ),
                _class="phanterpwa_datetimepicker_summary_current_data"
            ),
            _class="phanterpwa_datetimepicker_summary_container"
        )
        c_dt = ""
        if self.date_type == "date":
            c_dt = " phanterpwa_datetimepicker_is_not_datetime"

        datetimepicker_container = DIV(
            summary,
            container,
            _class="phanterpwa_datetimepicker_container{0}".format(c_dt)
        )

        centralizer = DIV(
            DIV(
                DIV(
                    DIV(
                        DIV(
                            datetimepicker_container,
                            _id="phanterpwa-centralizer-center-{0}".format(self.namespace),
                            _class="phanterpwa-centralizer-center"
                        ),
                        _class="phanterpwa-centralizer-horizontal"

                    ),
                    _class="phanterpwa-centralizer-vertical"
                ),
                _class="phanterpwa-centralizer-wrapper"
            ),
            _class="phanterpwa-fixed-fulldisplay",
            _style="z-index: 1008;"
        )

        def _selecting(el):
            v = jQuery(el).attr('phanterpwa_datetimepicker_date')
            self.set_selected(v)

        def open_panel(el):
            setTimeout(
                lambda: jQuery(self.target_selector).find(".phanterpwa-fixed-fulldisplay").addClass("enabled"),
                100
            )

        if jQuery(self.target_selector).has(".phanterpwa-fixed-fulldisplay").length == 1:
            # jQuery(self.target_selector).find(".phanterpwa-fixed-fulldisplay").remove()
            # # jQuery(self.target_selector).append(centralizer.xml())
            # centralizer.append_to(self.target_selector)
            datetimepicker_container.html_to("#phanterpwa-centralizer-center-{0}".format(self.namespace))
            jQuery(self.target_selector).find(".phanterpwa-fixed-fulldisplay").addClass("enabled")
        else:
            # jQuery(self.target_selector).append(centralizer.xml())
            centralizer.append_to(self.target_selector)
            open_panel()

        jQuery(
            '#phanterpwa_datetimepicker_prev_year_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_prev_year"
        ).on(
            "click.phanterpwa_datetimepicker_prev_year",
            lambda: self.previus_year()
        )
        jQuery(
            '#phanterpwa_datetimepicker_next_year_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_next_year"
        ).on(
            "click.phanterpwa_datetimepicker_next_year",
            lambda: self.next_year()
        )
        jQuery(
            '#phanterpwa_datetimepicker_prev_month_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_prev_month"
        ).on(
            "click.phanterpwa_datetimepicker_prev_month",
            lambda: self.previus_month()
        )
        jQuery(
            '#phanterpwa_datetimepicker_next_month_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_next_month"
        ).on(
            "click.phanterpwa_datetimepicker_next_month",
            lambda: self.next_month()
        )

        jQuery(
            '#phanterpwa_datetimepicker_prev_hour_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_prev_hour"
        ).on(
            "click.phanterpwa_datetimepicker_prev_hour",
            lambda: self.previus_hour()
        )
        jQuery(
            '#phanterpwa_datetimepicker_next_hour_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_next_hour"
        ).on(
            "click.phanterpwa_datetimepicker_next_hour",
            lambda: self.next_hour()
        )

        jQuery(
            '#phanterpwa_datetimepicker_prev_minute_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_prev_minute"
        ).on(
            "click.phanterpwa_datetimepicker_prev_minute",
            lambda: self.previus_minute()
        )
        jQuery(
            '#phanterpwa_datetimepicker_next_minute_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_next_minute"
        ).on(
            "click.phanterpwa_datetimepicker_next_minute",
            lambda: self.next_minute()
        )

        jQuery(
            '#phanterpwa_datetimepicker_prev_second_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_prev_second"
        ).on(
            "click.phanterpwa_datetimepicker_prev_second",
            lambda: self.previus_second()
        )
        jQuery(
            '#phanterpwa_datetimepicker_next_second_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_next_second"
        ).on(
            "click.phanterpwa_datetimepicker_next_second",
            lambda: self.next_second()
        )

        jQuery(
            '#phanterpwa_datetimepicker_month_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_month"
        ).on(
            "click.phanterpwa_datetimepicker_month",
            lambda: self.show_months()
        )
        jQuery(
            '#phanterpwa_datetimepicker_year_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_year"
        ).on(
            "click.phanterpwa_datetimepicker_year",
            lambda: self.show_years()
        )
        jQuery(
            '#phanterpwa_datetimepicker_hour_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_hour"
        ).on(
            "click.phanterpwa_datetimepicker_hour",
            lambda: self.show_hours()
        )

        jQuery(
            '#phanterpwa_datetimepicker_minute_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_minute"
        ).on(
            "click.phanterpwa_datetimepicker_minute",
            lambda: self.show_minutes()
        )

        jQuery(
            '#phanterpwa_datetimepicker_second_{0}'.format(self.namespace)
        ).off(
            "click.phanterpwa_datetimepicker_second"
        ).on(
            "click.phanterpwa_datetimepicker_second",
            lambda: self.show_seconds()
        )

        jQuery(
            '#phanterpwa_datetimepicker_calendar_{0} {1} .phanterpwa_datetimepicker_unit'.format(
                self.namespace, '.phanterpwa_datetimepicker_day'
            )
        ).off(
            'click.selecting_datepicker',
        ).on(
            'click.selecting_datepicker',
            lambda: _selecting(this)
        )
        jQuery(
            ".phanterpwa_datetimepicker_summary_container .phanterpwa-models-close"
        ).off(
            "click.deleting_datepicker",
        ).on(
            "click.deleting_datepicker",
            lambda: self.close()
        )


__pragma__('nokwargs')



project_config_sample = {
    "CONFIG_INDENTIFY": "project_config",
    "ENVIRONMENT": {
        "path": "",
        "python": ""
    },
    "PROJECT": {
        "name": "PhanterPWA",
        "title": "PhanterPWA",
        "description": "PhanterPWA Project",
        "version": "0.0.1",
        "compilation": 0,
        "author": "PhanterJR<phanterjr@conexaodidata.com.br>",
        "debug": True,
        "packaged": True,
        "minify": False,
        "baseport": 10000,
        "basehost": "127.0.0.1",
        "path": ""
    },
    "FRONTEND": {
        "sample_app": {
            "title": "PhanterPWA Frontend",
            "description": "Frontend Application PhanterPWA",
            "default_time_user_token_expire": 7200,
            "default_time_user_token_expire_remember_me": 2592000,
            "default_time_csrf_token_expire": 4200,
            "default_time_temporary_password_expire": 4200,
            "timeout_to_resend_temporary_password_mail": 3900,
            "default_time_client_token_expire": 63072000,
            "default_time_activation_code_expire": 3600,
            "wait_time_to_try_activate_again": 3900,
            "timeout_to_resend_activation_email": 300,
            "timeout_to_resign": 600,
            "timeout_to_next_login_attempt": 4200,
            "max_login_attempts": 5,
            "max_activation_attempts": 5,
            "default_time_user_token_expire": 7200,
            "default_time_two_factor_code_expire": 3600
        }
    },
    "BACKEND": {
        "sample_app": {
            "title": "PhanterPWA Backend",
            "description": "Backend Application PhanterPWA",
            "build_folder": "",
            "transcrypt_main_file": "application",
            "styles_main_file": "application",
            "views_main_file": "index"
        }
    }
}



import os
from tornado import (
    web
)
# from core import (
#     projectConfig,
#     Translator,
# )
from phanterpwa.i18n import (
    browser_language,
    Translator
)


class I18N(web.RequestHandler):
    def initialize(self, app_name, projectConfig, ignore_debug=True):
        self.app_name = app_name
        self.ignore_debug = ignore_debug
        self.projectConfig = projectConfig
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "phanterpwa-authorization,",
                "cache-control"
            ])
        )
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
        if self.request.headers.get("phanterpwa-language"):
            self.phanterpwa_language = self.request.headers.get("phanterpwa-language")
        else:
            self.phanterpwa_language = browser_language(self.request.headers.get("Accept-Language"))

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    def options(self, *args):
        self.set_status(200)
        return self.write({"status": "OK"})

    def get(self, *args):
        app = args[0]
        dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}
        lang = dict_arguments.get("lang", self.phanterpwa_language)
        word = dict_arguments.get("new_word", None)
        current_dir = self.projectConfig['PROJECT']['path']
        translator_instance = Translator(
            os.path.join(current_dir, "backapps", self.app_name, "languages", app),
            identifier=lang,
            debug=self.projectConfig['PROJECT']['debug']
        )

        if translator_instance:
            if word and lang and self.projectConfig["PROJECT"]["debug"]:
                translator_instance.add_language(lang)
                translator_instance.direct_translation = lang
                translator_instance.translator(word, lang)
            if self.projectConfig["PROJECT"]["debug"] or self.ignore_debug:
                return self.write(translator_instance.languages)
            else:
                return self.set_status(503)
        else:
            return self.set_status(404)



"""
Title: XmlConstructor

Author: PhanterJR<junior.conex@gmail.com>

License: MIT

Coding: utf-8

Send emails
"""
import requests
import json


class SMSSender(object):
    """
    
    """

    def __init__(self, fone_number, identify, text_message, app_name, projectConfig):
        super(SMSSender, self).__init__()
        self.fone_number = fone_number
        self.identify = identify
        self.text_message = text_message
        self.app_name = app_name
        self.projectConfig = projectConfig

    def send(self):
        """
            Mobizon Example
        """
        HEADERS = {
            'cache-control': 'no-cache',
            'content-type': 'application/x-www-form-urlencoded'
        }
        data = {
            'recipient': self.fone_number,
            'text': self.text_message
        }
        key = self.projectConfig["BACKEND"][self.app_name].get("mobizen_key", None)
        url = "https://api.mobizon.com.br/service/message/sendSmsMessage?output=json&api=v1&apiKey={0}".format(key)
        response_cl1 = requests.post(url, headers=HEADERS, data=data, timeout=30)
        json_response_cl1 = json.loads(response_cl1.content)
        return json_response_cl1



import phanterpwa.frontend.helpers as helpers
from org.transcrypt.stubs.browser import __pragma__


__pragma__('alias', "jQuery", "$")
__pragma__('skip')
jQuery = sessionStorage = JSON = js_undefined = setTimeout = window = this = console =\
    localStorage = 0
__pragma__('noskip')


DIV = helpers.XmlConstructor.tagger("div")
I = helpers.XmlConstructor.tagger("i")


__pragma__('kwargs')


class Pagination(object):
    """docstring for Pagination"""

    def __init__(self, target_element, **paramaters):
        self.target_element = jQuery(target_element)
        self._current_page = 1
        self._pages_list = [1]
        self._max_pages_on_screen = 11

    def start(self):
        html = DIV(
            _class="phanterpwa-pagination-wrapper"
        )


__pragma__('nokwargs')



import os
import io
import sys
import sass
import json
import importlib
import shutil
import subprocess
import phanterpwa
import time
import datetime
import re
import traceback
from pathlib import (PurePath, Path)
from glob import glob, iglob
from phanterpwa.configer import ProjectConfig
from phanterpwa.tools import interpolate
from os.path import (
    normpath,
    join,
    isdir,
    isfile,
    dirname,
    basename,
    exists,
    getmtime
)

ENV_PYTHON = normpath(sys.executable)
PATH_PHANTERPWA = dirname(phanterpwa.__file__)


ServiceWorker = """
self.addEventListener('install', function(event) {
  event.waitUntil(
    caches.open('{{versioning}}').then(function(cache) {
      return cache.addAll([
        {{files}}
      ]);
    })
  );
});

self.addEventListener('fetch', function(event) {
  if (event.request.method == "GET"){
      event.respondWith(caches.match(event.request).then(function(response) {
        if (response !== undefined) {
          return response;
        } else {
          return fetch(event.request).then(function (response) {
            return response;
          }).catch(function () {
            return caches.match('/static/{{versioning}}/images/warning.png');
          });
        }
      }));
  } else {
    if (window.PhanterPWA.DEBUG){
        console.info(event.request.method)
    };
  };
});
"""


class Compiler():
    """
        pt-BR: A classe Compilar, baseado na configuração do projeto (project.ini, api.ini e app.ini), axecuta as seguintes ações:
            Se a configuração do projeto estiver em debug: True
            - Converte os templates python pra o formato html
            -
    """

    def __init__(self, projectpath, build_apps_folder=None, reset_compilation=False):
        self.has_error = None
        self.ProjectConfig = ProjectConfig(projectpath)
        self.projectpath = projectpath
        self.config = self.ProjectConfig.config
        self._check_app_list()
        self.build_apps_folder = build_apps_folder
        if reset_compilation:
            self.current_compilation = 0

        else:
            self.current_compilation = self.config['PROJECT'].get("compilation", 0)
            self.current_compilation += 1
        self.config['PROJECT']["compilation"] = self.current_compilation
        self.config["PROJECT"]["date_compilation"] = str(datetime.datetime.now())
        self.config['PROJECT']['versioning'] = "{0}.{1}".format(
            self.config["PROJECT"].get("version", "0.0.1"),
            self.current_compilation
        )
        self.debug = self.config['PROJECT']["debug"]
        self.version = self.config['PROJECT']['version']
        self.versioning = self.config['PROJECT']['versioning']
        self.tempfolder = join(self.projectpath, "temp")
        if self.debug:
            self.minify = False
            self.full_compilation = False
            self.clear_builder_folder = False
        else:
            self.minify = True
            self.full_compilation = True
            self.clear_builder_folder = True
        self.ProjectConfig.save()
        self._templates_to_update = {}
        self._statics_to_update = {}
        self._styles_changed = {}
        self._phanterpwa_modules_changed = []
        self._has_config_changed = {}
        self._has_style_change = {}
        self._has_template_change = {}
        self._has_static_change = {}
        self._has_script_change = {}
        self._has_phanterpwa_modules_change = {}
        self._create_temp_folder()

    def _check_app_list(self):

        for app in list(self.config['FRONTEND'].keys()):
            if not isdir(join(self.projectpath, "frontapps", app)):
                del self.config['FRONTEND'][app]

    @property
    def app_list(self):
        """
            lista os aplicativos
        """
        self._check_app_list()
        return self.config['FRONTEND'].keys()

    @property
    def build_apps_folder(self):
        return self._build_apps_folder

    @build_apps_folder.setter
    def build_apps_folder(self, value):
        self._build_apps_folder = None
        if value is not None:
            if isdir(value):
                self._build_apps_folder = value
            else:
                raise IOError("The build apps folder don't found! Given: {0}".format(value))

    @property
    def minify(self):
        return self._minify

    @minify.setter
    def minify(self, value):
        if isinstance(value, bool):
            self._minify = value
        else:
            raise ValueError("The minify attribute must be boolean type. Given: {0}".format(type(value)))

    @property
    def projectpath(self):
        return self._projectpath

    @projectpath.setter
    def projectpath(self, value):
        if isdir(join(value)):
            if isfile(join(value, "config.json")):
                self._projectpath = value
            else:
                raise ValueError("the project path don't have the config file (config.json).")
        else:
            raise IOError("The project path is invalid!. Given: {0}".format(value))

    @staticmethod
    def get_files_dir(path, ignore_files=[], ignore_paths=[]):
        r = os.walk(path)
        for x in r:
            if basename(x[0]) not in ignore_paths:
                for y in x[2]:
                    if basename(y) not in ignore_files:
                        yield normpath(join(x[0], y))
        pass

    def templates_files(self, app, include_extends=False):
        if include_extends:
            ig_paths = ["__pycache__"]
        else:
            ig_paths = ["__pycache__", "extends"]
        files = self.get_files_dir(
            self.path_templates_folder(app), ignore_files=["__init__.py"], ignore_paths=ig_paths)
        return (x for x in files if x.endswith(".py"))

    def extends_files(self, app):
        files = self.get_files_dir(
            join(self.path_templates_folder(app), "extends"),
                ignore_files=["__init__.py"], ignore_paths=["__pycache__"])
        return (x for x in files if x.endswith(".py"))

    def static_files(self, app):
        files = self.get_files_dir(
            self.path_statics_folder(app), ignore_files=["__init__.py"], ignore_paths=["__pycache__"])
        return (x for x in files)

    def transcrypt_files(self, app):
        files = self.get_files_dir(
            self.path_transcrypts_folder(app), ignore_files=["__init__.py"],
                ignore_paths=["__pycache__", "__target__"])
        return (x for x in files if x.endswith(".py"))

    def style_files(self, app):
        files = self.get_files_dir(
            self.path_styles_folder(app), ignore_files=["__init__.py"], ignore_paths=["__pycache__"])
        return (normpath(x) for x in files if x.endswith(".sass"))

    def phanterpwa_usual_sass(self, app):
        print("\nProcess PhanterPWA usual sass...")
        variables = join(PATH_PHANTERPWA, "usual_sass", "variables.sass")
        main_file = join(PATH_PHANTERPWA, "usual_sass", "phanterpwa.sass")
        # if exists(join(dirname(main_file), "_compiler_sass_temp_file.sass")):
        #     os.unlink(join(dirname(main_file), "_compiler_sass_temp_file.sass"))
        target_dir = join(self.projectpath, "frontapps", app, "statics", "css")
        if not isdir(target_dir):
            try:
                os.makedirs(target_dir, exist_ok=True)
            except OSError as e:
                raise e("Problem on create folder '{0}'.".format(target_dir))
        target_css = join(target_dir, "phanterpwa.css")

        sfiles = self.get_files_dir(
            join(PATH_PHANTERPWA, "usual_sass"), ignore_files=["__init__.py"], ignore_paths=["__pycache__"])
        sass_files = (normpath(x) for x in sfiles if x.endswith(".sass"))

        with open(variables, 'r', encoding="utf-8") as f:
            txt = f.read()

        new_text_to_save = "".join([
            "//change programmatically\n",
            "$app-version: \"{0}\"\n".format(self.versioning),
            "//change programmatically\n"
        ])
        with open(main_file, 'r', encoding="utf-8") as f:
            main_txt = f.read()
        txt = "\n".join([
            new_text_to_save,
            "\n/* SASS Source Code (VARIABLES): {0} */".format(variables),
            txt,
            "\n/* SASS Source Code (MAIN FILE): {0} */".format(main_file),
            main_txt
        ])
        has_import = False
        for x in sass_files:
            if x != normpath(main_file) and x !=normpath(variables):
                c = "/* SASS Source Code: {0} */\n\n".format(x)
                p = PurePath(x)
                p = p.relative_to(dirname(main_file))
                l = [*p.parts]
                l[-1] = l[-1][0:-5]
                patter = "@import '{0}'".format("/".join(l))
                txt = "".join(["\n", txt, c, patter, "\n\n"])
                has_import = True
        if has_import:
            print("    Creating imports...")
        # ftempsass = io.StringIO()
        # with open(ftempsass, "w", encoding="utf-8") as f2:
        #     f2.write(txt)
        print("    Convert sass to css: {0}".format(target_css))
        if self.minify:
            new_css = sass.compile(
                string=txt,
                output_style="compressed",
                indented=True,
                include_paths=[dirname(main_file)]
            )
        else:
            new_css = sass.compile(
                string=txt,
                output_style="expanded",
                indented=True,
                include_paths=[dirname(main_file)]
            )
        # if isfile(join(self.projectpath, "temp", "_compiler_sass_temp_file.sass")):
        #     os.unlink(join(self.projectpath, "temp", "_compiler_sass_temp_file.sass"))
        # shutil.copy(
        #     join(dirname(main_file), "_compiler_sass_temp_file.sass"),
        #     join(self.projectpath, "temp", "_compiler_sass_temp_file.sass")
        # )
        with open(target_css, "w") as o:
            o.write(new_css)
        # os.unlink(join(dirname(main_file), "_compiler_sass_temp_file.sass"))
        # with open(main_file, "w", encoding="utf-8") as fw:
        #     fw.write(new_text_to_save)

    @staticmethod
    def target_by_relative_path(src_path, tgt_path, ext_src=None, ext_tgt=None, ignore_files=[], ignore_paths=[]):

        def _target(file_path, relative_to, tgt_path, ext_src, ext_tgt):
            p = PurePath(file_path)
            p = p.relative_to(relative_to)
            target_file = join(tgt_path, *p.parts)
            if ext_tgt is not None:
                target_file = "".join([target_file[0:-len(ext_src)], ext_tgt])
            return [file_path, target_file]

        if ext_tgt is not None:
            if ext_src is None:
                raise SyntaxError("The ext_src can't be None when ext_tgt is different of None.")
            if not ext_tgt.startswith("."):
                raise ValueError("The ext_tgt must be starts with dot(.) or must be None. Examplo: .py")

        if ext_src is not None:
            if not ext_src.startswith("."):
                raise ValueError("The ext_src must be starts with dot(.) or must be None. Examplo: .py")
            return (
                _target(
                    f,
                    relative_to=src_path,
                    tgt_path=tgt_path, ext_src=ext_src,
                    ext_tgt=ext_tgt
                ) for f in Compiler.get_files_dir(src_path,
                        ignore_files=ignore_files,
                        ignore_paths=ignore_paths) if f.endswith(ext_src)
            )
        else:
            return (
                _target(
                    f,
                    relative_to=src_path,
                    tgt_path=tgt_path, ext_src=ext_src,
                    ext_tgt=ext_tgt
                ) for f in Compiler.get_files_dir(src_path,
                        ignore_files=ignore_files,
                        ignore_paths=ignore_paths)
            )

    def _create_temp_folder(self):
        try:
            os.makedirs(join(self.tempfolder), exist_ok=True)
        except OSError as e:
            raise e("Problem on create folder '{0}'.".format(join(self.tempfolder)))

    def path_build_templates_folder(self, app):
        if self.build_apps_folder:
            return join(self.build_apps_folder, app, "www")
        else:
            return self.config["FRONTEND"][app]["build_folder"]

    def path_build_statics_folder(self, app):
        if self.build_apps_folder:
            return join(self.build_apps_folder, app, "www", "static")
        else:
            return join(self.config["FRONTEND"][app]["build_folder"], "static")

    def path_build_styles_folder(self, app):
        if self.build_apps_folder:
            return join(self.build_apps_folder, app, "www", "static", self.versioning, "css")
        else:
            return join(self.config["FRONTEND"][app]["build_folder"], "static", self.versioning, "css")

    def path_build_transcrypt_folder(self, app):
        if self.build_apps_folder:
            return join(self.build_apps_folder, app, "www", "static", self.versioning, "js", "transcrypt")
        else:
            return join(self.config["FRONTEND"][app]["build_folder"], "static", self.versioning, "js", "transcrypt")

    def path_templates_folder(self, app):
        return join(self.projectpath, "frontapps", app, "sources", "templates")

    def path_styles_folder(self, app):
        return join(self.projectpath, "frontapps", app, "sources", "styles")

    def path_transcrypts_folder(self, app):
        return join(self.projectpath, "frontapps", app, "sources", "transcrypts")

    def path_statics_folder(self, app):
        return join(self.projectpath, "frontapps", app, "statics")

    def path_app_config_file(self, app):
        os.makedirs(join(self.path_transcrypts_folder(app), "auto"), exist_ok=True)
        return join(self.path_transcrypts_folder(app), "auto", "config.py")

    def target_template_file_by_source(self, src, app) -> str:
        relative_to = self.path_templates_folder(app)
        target_path = self.path_build_templates_folder(app)
        p = PurePath(src)
        p = p.relative_to(relative_to)
        target_file = join(target_path, *p.parts)
        target_file = "".join([target_file[0:-3], ".html"])
        return target_file

    def _check_templates_target_exist(self, src, app) -> bool:
        target = self.target_template_file_by_source(src, app)
        return isfile(target)

    def target_static_file_by_source(self, src, app) -> str:
        relative_to = self.path_statics_folder(app)
        target_path = join(self.path_build_statics_folder(app), self.versioning)
        p = PurePath(src)
        p = p.relative_to(relative_to)
        target_file = join(target_path, *p.parts)
        return target_file

    def _check_statics_target_exist(self, src, app) -> bool:
        target = self.target_static_file_by_source(src, app)
        return isfile(target)

    def _check_style_target_exist(self, app):
        target = join(self.path_build_styles_folder(app),
            "{0}.css".format(self.config["FRONTEND"][app]["styles_main_file"]))
        return isfile(target)

    def _check_transcrypt_target_exist(self, app):
        target = join(self.path_build_transcrypt_folder(app),
            "{0}.js".format(self.config["FRONTEND"][app]["transcrypt_main_file"]))
        return isfile(target)

    def create_templates_to_update_generator(self, app):
        if any([self.full_compilation, self._has_config_changed.get(app, False),
                self._check_template_extends_change(app)]):
            self._templates_to_update[app] = self.templates_files(app)
        else:
            self._templates_to_update[app] = []
            t_files = self.templates_files(app)
            if isfile(join(self.tempfolder, "templates_mtime_{0}.json".format(app))):
                with open(join(self.tempfolder,
                        "templates_mtime_{0}.json".format(app)), "r", encoding="utf-8") as f:
                    content = json.load(f)
                    for x in t_files:
                        if x in content:
                            if getmtime(x) != content[x] or not self._check_templates_target_exist(x, app):
                                self._templates_to_update[app].append(x)
                        else:
                            self._templates_to_update[app].append(x)
            else:
                self._templates_to_update[app] = self.templates_files(app)

    def create_statics_to_update_generator(self, app):
        if any([self.full_compilation, self._has_config_changed.get(app, False)]):
            self._statics_to_update[app] = self.static_files(app)
        else:
            self._statics_to_update[app] = []
            t_files = self.static_files(app)
            if isfile(join(self.tempfolder, "statics_mtime_{0}.json".format(app))):
                with open(join(self.tempfolder,
                        "statics_mtime_{0}.json".format(app)), "r", encoding="utf-8") as f:
                    content = json.load(f)
                    for x in t_files:
                        if x in content:
                            if getmtime(x) != content[x] or not self._check_statics_target_exist(x, app):
                                self._statics_to_update[app].append(x)
                        else:
                            self._statics_to_update[app].append(x)
            else:
                self._statics_to_update[app] = self.static_files(app)

    def _check_styles_change(self, app):
        if any([self.full_compilation,
                self._has_config_changed.get(app, False), not self._check_style_target_exist(app)]):
            return True
        t_files = self.style_files(app)
        if isfile(join(self.tempfolder, "styles_mtime_{0}.json".format(app))):
            with open(join(self.tempfolder,
                    "styles_mtime_{0}.json".format(app)), "r", encoding="utf-8") as f:
                content = json.load(f)
                for x in t_files:
                    if x in content:
                        if getmtime(x) != content[x]:
                            return True
                    else:
                        return True
        else:
            return True
        return False

    def _check_script_change(self, app):
        if any([self.full_compilation,
                self._has_config_changed.get(app, False), not self._check_transcrypt_target_exist(app)]):
            return True
        t_files = self.transcrypt_files(app)
        if isfile(join(self.tempfolder, "transcrypts_mtime_{0}.json".format(app))):
            with open(join(self.tempfolder,
                    "transcrypts_mtime_{0}.json".format(app)), "r", encoding="utf-8") as f:
                content = json.load(f)
                for x in t_files:
                    if x in content:
                        if getmtime(x) != content[x]:
                            return True
                    else:
                        return True
                for x in content:
                    if not isfile(x):
                        return True
        else:
            return True
        return False

    def _check_template_extends_change(self, app):
        extends_files = self.extends_files(app)
        if isfile(join(self.tempfolder, "extends_mtime_{0}.json".format(app))):
            with open(join(self.tempfolder,
                    "extends_mtime_{0}.json".format(app)), "r", encoding="utf-8") as f:
                content = json.load(f)
                for x in extends_files:
                    if x in content:
                        if getmtime(x) != content[x]:
                            return True
                    else:
                        return True
                return False
        else:
            return True

    def copy_statics(self, app) -> list:
        print("\nProcess statics...")
        self.create_statics_to_update_generator(app)
        if self.full_compilation:
            path_build_statics_folder_versioned = join(self.path_build_statics_folder(app), self.versioning)
            path_statics_folder = self.path_statics_folder(app)
            if exists(path_build_statics_folder_versioned) and isdir(path_build_statics_folder_versioned):
                shutil.rmtree(path_build_statics_folder_versioned)
            if exists(path_statics_folder):
                shutil.copytree(
                    path_statics_folder,
                    path_build_statics_folder_versioned
                )
                print("    Coping on", join(path_build_statics_folder_versioned))
        else:
            paths_on_static_folder = (x for x in glob(join(self.path_build_statics_folder(app), "*")) if isdir(x))
            for pat in paths_on_static_folder:
                if basename(pat) != self.versioning:
                    try:
                        shutil.rmtree(pat)
                    except Exception as e:
                        print(e)
            has_copy = False
            for x in self._statics_to_update[app]:
                rel = self.target_static_file_by_source(x, app)
                print("    Coping statics: {0} ---> {1}".format(x, rel))
                try:
                    os.makedirs(dirname(rel), exist_ok=True)
                except OSError as e:
                    raise e("Problem on create folder '{0}'.".format(dirname(rel)))
                shutil.copy2(x, rel)
                has_copy = True
            if not has_copy:
                pass
                print("    Skiping Static...")

    def delete_compiled_app_folder(self, app):
        appConfig = self.config
        if self.full_compilation:
            target = appConfig.get('FRONTEND')[app]['build_folder']
            if target:
                if exists(target) and isdir(target):
                    print("Deleting builder folder...")
                    try:
                        shutil.rmtree(target)
                    except OSError as e:
                        print("Error: %s - %s." % (e.filename, e.strerror))
                        pass
                    while exists(target):
                        pass
                    print("Deleted ({0})".format(target))

    def copy_languages(self, app):
        print("\nCopying languages...")
        appConfig = self.config
        version = self.versioning
        apps_list_basedir = join(appConfig['PROJECT']['path'], "frontapps")
        source_apps = join(
            apps_list_basedir,
            "languages"
        )
        folder_lang_apps_list = join(
            appConfig["FRONTEND"][app]['build_folder'],
            "static",
            version,
            "languages"
        )
        os.makedirs(
            join(
                folder_lang_apps_list), exist_ok=True)
        if not exists(folder_lang_apps_list):
            os.makedirs(
                join(folder_lang_apps_list), exist_ok=True
            )
        langs = glob(join(source_apps, "*.json"))
        for y in langs:
            with open(y, "r", encoding='utf-8') as f:
                c = json.load(f)
                lang_file = join(folder_lang_apps_list, basename(y))
                with open(lang_file, "w", encoding='utf-8') as o:
                    if self.minify:
                        json.dump(c, o, ensure_ascii=False)
                    else:
                        json.dump(c, o, ensure_ascii=False, indent=2)

    def compile_styles(self, app):
        print("\nProcess styles...")
        appConfig = self.config
        changed = self._check_styles_change(app)
        if changed or self.full_compilation:
            variables = join(PATH_PHANTERPWA, "usual_sass", "variables.sass")
            with open(variables, 'r', encoding="utf-8") as f:
                txt = f.read()
            main_file = join(
                self.path_styles_folder(app), "{0}.sass".format(
                    appConfig["FRONTEND"][app]['styles_main_file'])
            )
            if not exists(main_file):
                Path(main_file).touch()

            # if exists(join(dirname(main_file), "_compiler_sass_temp_file.sass")):
            #     os.unlink(join(dirname(main_file), "_compiler_sass_temp_file.sass"))
            target_css = join(
                self.path_build_styles_folder(app),
                    "{0}.css".format(appConfig["FRONTEND"][app]['styles_main_file'])
            )
            try:
                os.makedirs(dirname(target_css), exist_ok=True)
            except OSError as e:
                raise e("Problem on create folder '{0}'.".format(target_css))

            sass_files = self.style_files(app)

            variables = join(PATH_PHANTERPWA, "usual_sass", "variables.sass")
            with open(variables, 'r', encoding="utf-8") as f:
                txt = f.read()

            new_text_to_save = "".join([
                "//change programmatically\n",
                "$app-version: \"{0}\"\n".format(self.versioning),
                "//change programmatically\n"
            ])
            with open(main_file, 'r', encoding="utf-8") as f:
                main_txt = f.read()
            txt = "\n".join([
                new_text_to_save,
                "\n/* SASS Source Code (VARIABLES): {0} */".format(variables),
                txt,
                "\n/* SASS Source Code (MAIN FILE): {0} */".format(main_file),
                main_txt
            ])
            has_import = False
            for x in sass_files:
                if x != normpath(main_file):
                    c = "/* SASS Source Code: {0} */\n\n".format(x)
                    p = PurePath(x)
                    p = p.relative_to(dirname(main_file))
                    l = [*p.parts]
                    l[-1] = l[-1][0:-5]
                    patter = "@import '{0}'".format("/".join(l))
                    txt = "".join(["\n", txt, c, patter, "\n\n"])
                    has_import = True
            if has_import:
                print("    Creating imports")

            # with open(join(dirname(main_file), "_compiler_sass_temp_file.sass"), "w", encoding="utf-8") as f2:
            #     f2.write(txt)
            print("    Convert sass to css: {0}".format(target_css))
            if self.minify:
                new_css = sass.compile(
                    string=txt,
                    output_style="compressed",
                    indented=True,
                    include_paths=[dirname(main_file)]
                )
            else:
                new_css = sass.compile(
                    string=txt,
                    output_style="expanded",
                    indented=True,
                    include_paths=[dirname(main_file)]
                )
            # if isfile(join(self.projectpath, "temp", "_compiler_sass_temp_file.sass")):
            #     os.unlink(join(self.projectpath, "temp", "_compiler_sass_temp_file.sass"))
            # shutil.copy(
            #     join(dirname(main_file), "_compiler_sass_temp_file.sass"),
            #     join(self.projectpath, "temp", "_compiler_sass_temp_file.sass")
            # )
            with open(target_css, "w") as o:
                o.write(new_css)
            # os.unlink(join(dirname(main_file), "_compiler_sass_temp_file.sass"))
            # with open(main_file, "w", encoding="utf-8") as fw:
            #     fw.write(new_text_to_save)
        else:
            print("    Skiping styles...")

    def compile_templates(self, app):
        print("\nProcess Templates...")
        self.create_templates_to_update_generator(app)
        project_path = self.projectpath
        appConfig = self.config
        sys.path.append(project_path)
        os.chdir(project_path)
        base_dir = join(appConfig['PROJECT']['path'], "frontapps", app)

        def _compile_html(file, base="", target=None, is_apps=False, app_name="", ignore=["__init__.py"]):

            if isfile(file) and basename(file) not in ignore and file[-3:] == ".py":
                i_mod = "%s" % (basename(file)[0:-3])
                if base:
                    i_mod = "%s.%s" % (base, i_mod)
                i = importlib.import_module(i_mod)
                importlib.reload(i)
                name = "".join([*i_mod.split(".")[-1], ".html"])
                f_parts = i_mod.split(".")[3:-1]
                files_www = join(target, *f_parts)
                if is_apps:
                    files_www = join(target, *f_parts[1:])
                if not exists(join(files_www)):
                    try:
                        os.makedirs(join(files_www), exist_ok=True)
                    except OSError as e:
                        raise e("Problem on create folder '{0}'.".format(join(files_www)))
                print("    Convert python to html: {0} ---> {1})".format(file, join(files_www, name)))
                with open(
                    join(files_www, name),
                    "wt",
                    encoding="utf-8"
                ) as f:
                    if self.minify:
                        f.write(i.html.xml())
                    else:
                        f.write(i.html.humanize())

        def _compile_htmls(source, base="", target=None, is_apps=False, app_name=""):
            list_all = iglob(join(source, "*"))
            ignore_paths = ["__pycache__", "extends"]
            for x in list_all:
                bname = basename(x)
                if isdir(x) and bname not in ignore_paths:
                    if base:
                        new_base = "{0}.{1}".format(base, basename(x))
                        _compile_htmls(x, new_base, target=target, is_apps=is_apps, app_name=app_name)
                elif isfile(x) and x in self._templates_to_update[app]:
                    _compile_html(x, base=base, target=target, is_apps=is_apps, app_name=app_name)

        target_apps = self.path_build_templates_folder(app)
        self._templates_to_update[app] = [x for x in self._templates_to_update[app]]
        if self._templates_to_update[app]:
            _compile_htmls(
                join(base_dir, "sources", "templates"),
                "frontapps.{0}.sources.templates".format(app),
                target=target_apps,
                is_apps=True,
                app_name=app
            )
        else:
            print("    Skiping templates...")
            self._cont_templates = 0

    def compile_transcrypts(self, app):
        print("\nProcess scrypts...")
        appConfig = self.config
        changed = self._check_script_change(app)
        if not (changed or self.full_compilation):
            print("    Local files no changed...")
        python_env = appConfig['ENVIRONMENT']['python']
        folder_script_apps_list = self.path_build_transcrypt_folder(app)
        os.makedirs(
            join(
                folder_script_apps_list), exist_ok=True)
        source = join(self.path_transcrypts_folder(app), "__target__")
        files_trans = []
        for t in glob(join(self.path_transcrypts_folder(app), "*.py")):
            if not basename(t).startswith("_"):
                files_trans.append(t)
        for main_file in files_trans:
            now = datetime.datetime.now()
            mdtime = time.mktime(now.timetuple())
            os.utime(main_file, (mdtime, mdtime))
            print("    Convert python to javascript: {0}".format(main_file))
            print("    Starting Transcrypt compiler. Wait to complete.")
            if self.minify:
                print("    For minification it's necessary to have java installed, if compilation fail,",
                    " the compilation will try transcrypt on unminify format.")
                try:
                    subprocess.run("{0} -X utf8 -m transcrypt {1}".format(python_env, main_file), shell=True, check=True)
                except Exception as e:
                    print("    Minification Fail!!! It's try unminify mode now, it's fast.",
                        " Check java instalation. Error:", e)
                    subprocess.run("{0} -X utf8 -m transcrypt {1} -n".format(python_env, main_file), shell=True, check=True)
            else:
                subprocess.run("{0} -X utf8 -m transcrypt {1} -n -m".format(python_env, main_file), shell=True, check=True)
            list_all = glob(join(source, "*"))
            print("Coping scripts from __target__ to {0}\n".format(folder_script_apps_list))
            for y in list_all:
                if isfile(y):
                    script_file = join(
                        folder_script_apps_list, basename(y)
                    )
                    shutil.copy(
                        y,
                        script_file
                    )
                    if self.minify and not self.debug:
                        with open(script_file, "r", encoding="utf-8") as f:
                            lines_script = f.readlines()
                        if lines_script:
                            if lines_script[-1].startswith("//# sourceMappingURL"):
                                with open(script_file, "w", encoding="utf-8") as f:
                                    f.write("".join(lines_script[0:-1]))

        print("Finish!!!\n\n\n")

    def compile_sw(self, app):
        appConfig = self.config
        build_folder = appConfig.get('FRONTEND')[app]['build_folder']
        files = self.get_files_dir(
            build_folder, ignore_files=["__init__.py"])
        str_files = ""
        cont = 0
        for x in files:
            if x.startswith(join(build_folder, "static")):
                p = PurePath(x)
                p = p.relative_to(build_folder)
                l = [*p.parts]
                y = "/".join(l)
                if not cont:
                    str_files = "'/{0}'".format(y)
                else:
                    str_files = ",".join([str_files, "\n        '/{0}'".format(y)])
                cont = 1

        sw = interpolate(ServiceWorker, {
            "files": str(str_files),
            "versioning": self.versioning
        })
        with open(join(build_folder, 'sw.js'), "w") as o:
            o.write(sw)

    def transcrypts_config(self, app=None):
        if app:
            self._has_config_changed[app] = self._process_transcrypt_config(app)
        else:
            for app in self.app_list:
                self._has_config_changed[app] = self._process_transcrypt_config(app)

    def _process_transcrypt_config(self, app):
        ignore_keys = [
            "transcrypt_main_file"
            "styles_main_file",
            "views_main_file",
            "host",
            "port"
        ]

        last_app_config = join(self.tempfolder, "project_config_{0}.json".format(app))
        path_app_config_file = self.path_app_config_file(app)
        CONFIG = {
            "PROJECT": {
                "name": self.config["PROJECT"]["name"],
                "title": self.config["PROJECT"]["title"],
                "version": self.config["PROJECT"]["version"],
                "compilation": self.config["PROJECT"]["compilation"],
                "date_compilation": self.config["PROJECT"]["date_compilation"],
                "versioning": self.config["PROJECT"]["versioning"],
                "debug": self.config["PROJECT"]["debug"],
                "author": self.config["PROJECT"]["author"],
            },
            "APP": {
            },
            "I18N": {
                "default_lang": self.config["PROJECT"].get("default_language", None)

            }
        }

        social_logins = {}
        if "OAUTH_GOOGLE" in self.config:
            social_logins = {
                    "google": {
                        "remote_address": self.config['OAUTH_GOOGLE'].get("remote_address", ""),
                        "client_id": self.config['OAUTH_GOOGLE'].get("client_id", "")
                }
            }

        if social_logins:
            CONFIG["SOCIAL_LOGINS"] = social_logins

        for x in self.config['FRONTEND'][app]:
            if x not in ignore_keys:
                CONFIG['APP'][x] = self.config['FRONTEND'][app][x]

        if "SMS" in self.config:
            SMS = {
                "prefix": self.config['SMS'].get("prefix", "55"),
                "mask": self.config['SMS'].get("mask", "(##) # ####-####")
            }
            CONFIG["SMS"] = SMS

        CONFIG['APP']['name'] = app

        ini = "\n".join([
            "# Created automatically.",
            "#",
            "# In development it may be necessary to add static data",
            "# to the client side application after compiling, use",
            "# the CONFIGJS section of the application's app.ini",
            "# file for this.",
            "#\n\n",
        ])
        i18n_files = join(self.projectpath, "frontapps", app, "languages")
        files = glob(join(i18n_files, "*.json"))
        i18n_languages = {"default_lang": self.config["PROJECT"].get("default_language", None)}
        for x in files:
            if basename(x) != "entries.json":
                lang = basename(x)[0:-5]
                with open(x, encoding="utf-8") as f:
                    translates = json.load(f)
                i18n_languages[lang] = translates
        CONFIG['I18N'] = i18n_languages
        end = "\n"
        content = "".join([ini, "CONFIG = {0}".format(json.dumps(CONFIG, ensure_ascii=True, indent=4)), end])
        content = re.sub(r"(:[\t\n\r ]{0,}true[\t\n\r ]{0,})[,}]", ": True,", content)
        content = re.sub(r"(:[\t\n\r ]{0,}false[\t\n\r ]{0,})[,}]", ": False,", content)
        content = re.sub(r"(:[\t\n\r ]{0,}null[\t\n\r ]{0,})[,}]", ": None,", content)
        with open(path_app_config_file, "w", encoding="utf-8") as f:
            f.write(content)
        if not isfile(last_app_config):
            with open(last_app_config, "w", encoding="utf-8") as f:
                new_c = json.dumps(self.config)
                new_c = json.loads(new_c)
                new_c["FRONTEND"] = {app: new_c["FRONTEND"][app]}
                json.dump(new_c, f, ensure_ascii=True, indent=2)
            return True
        else:
            with open(last_app_config, "r", encoding="utf-8") as f:
                v = json.load(f)
                new_v = v
                new_v['PROJECT']['compilation'] = self.config['PROJECT']['compilation']
                new_c = json.dumps(self.config, ensure_ascii=True)
                new_c = json.loads(new_c)
                new_c["FRONTEND"] = {app: new_c["FRONTEND"][app]}
                new_v["FRONTEND"] = {app: new_v["FRONTEND"][app]}

                if new_v == new_c:
                    if exists(join(self.tempfolder, "transcrypts_mtime_{0}.json".format(app))):
                        with open(join(self.tempfolder,
                                "transcrypts_mtime_{0}.json".format(app)), "r", encoding="utf-8") as f:
                            changes_register = json.load(f)
                        if path_app_config_file in changes_register:
                            changes_register[path_app_config_file] = getmtime(path_app_config_file)
                        with open(join(self.tempfolder,
                                "transcrypts_mtime_{0}.json".format(app)), "w", encoding="utf-8") as f:
                            json.dump(changes_register, f, ensure_ascii=True, indent=2)
                    return False
            with open(last_app_config, "w", encoding="utf-8") as f:
                new_c = json.dumps(self.config)
                new_c = json.loads(new_c)
                new_c["FRONTEND"] = {app: new_c["FRONTEND"][app]}
                json.dump(new_c, f, ensure_ascii=True, indent=2)
            return True

    def _save_mtimes(self):
        for app in self.app_list:
            content = {x: getmtime(x) for x in self.style_files(app)}
            with open(join(self.tempfolder,
                    "styles_mtime_{0}.json".format(app)), "w", encoding="utf-8") as f:
                json.dump(content, f, ensure_ascii=True, indent=2)

            content = {x: getmtime(x) for x in self.extends_files(app)}
            with open(join(self.tempfolder,
                    "extends_mtime_{0}.json".format(app)), "w", encoding="utf-8") as f:
                if isinstance(content, dict):
                    json.dump(content, f, ensure_ascii=True, indent=2)
                else:
                    raise "The templates_mtime content must be dict type. Given: {0}".format(content)

            content = {x: getmtime(x) for x in self.templates_files(app)}
            with open(join(self.tempfolder,
                    "templates_mtime_{0}.json".format(app)), "w", encoding="utf-8") as f:
                if isinstance(content, dict):
                    json.dump(content, f, ensure_ascii=True, indent=2)
                else:
                    raise "The templates_mtime content must be dict type. Given: {0}".format(content)

            content = content = {x: getmtime(x) for x in self.static_files(app)}
            with open(join(self.tempfolder,
                    "statics_mtime_{0}.json".format(app)), "w", encoding="utf-8") as f:
                if isinstance(content, dict):
                    json.dump(content, f, ensure_ascii=True, indent=2)
                else:
                    raise "The statics_mtime content must be dict type. Given: {0}".format(content)

            content = {x: getmtime(x) for x in self.transcrypt_files(app)}
            with open(join(self.tempfolder,
                    "transcrypts_mtime_{0}.json".format(app)), "w", encoding="utf-8") as f:
                if isinstance(content, dict):
                    json.dump(content, f, ensure_ascii=True, indent=2)
                else:
                    raise "The scripts_mtime content must be dict type. Given: {0}".format(content)

            # content = {x: getmtime(x) for x in self.modules_files_monitor()}
            # with open(join(self.tempfolder,
            #         "phanterpwa_modules_mtime.json"), "w", encoding="utf-8") as f:
            #     json.dump(content, f, ensure_ascii=True, indent=2)

    def compile_by_step(self, app=None, full_compilation=False, minify=False):
        app_list = self.app_list
        if app:
            app_list = [app]
        for app in app_list:
            msg = "Transcrypt config"
            try:
                self.transcrypt_config(app)
                yield [msg, True, "Pass"]
            except Exception as e:
                yield [msg, False, traceback.format_tb(e.__traceback__)]
            current_debug = self.config["PROJECT"]["debug"]
            self.minify = self.config['PROJECT'].get("minify", False)
            # if current_debug:
            #     self.full_compilation = False
            # else:
            #     self.full_compilation = True
            # self._check_phanterpwa_modules()
            if full_compilation is not None:
                self.full_compilation = bool(full_compilation)
            if minify is not None:
                self.minify = bool(minify)
            msg = "Delete compiled app folder"
            try:
                self.delete_compiled_app_folder(app)
                yield [msg, True, "Pass"]
            except Exception as e:
                yield [msg, True, "".join(traceback.format_tb(e.__traceback__))]
            if current_debug or not exists(join(self.projectpath, "frontapps", app, "statics", "css", "phanterpwa.css")):
                msg = "Creating phanterpwa.css"
                try:
                    self.phanterpwa_usual_sass(app)
                    yield [msg, True, "Pass"]
                except Exception as e:
                    yield [msg, True, "".join(traceback.format_tb(e.__traceback__))]

            msg = "Copy statics"
            try:
                self.copy_statics(app)
                yield [msg, True, "Pass"]
            except Exception as e:
                yield [msg, True, "".join(traceback.format_tb(e.__traceback__))]
            msg = "Copy languages"
            try:
                self.copy_languages(app)
                yield [msg, True, "Pass"]
            except Exception as e:
                yield [msg, True, "".join(traceback.format_tb(e.__traceback__))]
            msg = "Compile styles"
            try:
                self.compile_styles(app)
                yield [msg, True, "Pass"]
            except Exception as e:
                yield [msg, True, "".join(traceback.format_tb(e.__traceback__))]
            msg = "Compile templates"
            try:
                self.compile_templates(app)
                yield [msg, True, "Pass"]
            except Exception as e:
                yield [msg, True, "".join(traceback.format_tb(e.__traceback__))]
            msg = "Compile transcrypts"
            try:
                self.compile_transcrypts(app)
                yield [msg, True, "Pass"]
            except Exception as e:
                yield [msg, True, "".join(traceback.format_tb(e.__traceback__))]
        self._save_mtimes()

    def compile(self,
        app=None, full_compilation=None,
        minify=None,
        force_transcrypt_compilation=True,
        ):

        app_list = self.app_list
        if app:
            app_list = [app]
        for app in app_list:
            self.transcrypts_config(app)
            print("\n============ APP COMPILATION: {0} ==============".format(app))
            current_debug = self.config["PROJECT"]["debug"]
            self.minify = self.config["PROJECT"].get("minify", False)
            print("DEBUG:", current_debug, "\t\t MINIFY:", self.minify)
            # if current_debug:
            #     self.full_compilation = False
            # else:
            #     self.full_compilation = True
            # self._check_phanterpwa_modules()
            if full_compilation is not None:
                self.full_compilation = bool(full_compilation)
            if minify is not None:
                self.minify = bool(minify)
            self.delete_compiled_app_folder(app)
            self.phanterpwa_usual_sass(app)
            self.copy_languages(app)
            self.compile_templates(app)
            self.copy_statics(app)
            self.compile_styles(app)
            self.compile_transcrypts(app)
            self.compile_sw(app)
        self._save_mtimes()
